# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
import scipy.constants as codata
from abc import abstractmethod
from pySRU.MagneticField import MagneticField

PLANE_UNDULATOR=0
BENDING_MAGNET=1

class MagneticStructure(object):
    def __init__(self, magnet_type ):
        self.magnet_type=magnet_type

    @abstractmethod
    def copy(self):
        return


    @abstractmethod
    def print_parameters(self):
        pass


    # magnetic field
    @abstractmethod
    def fct_magnetic_field(self, z, y, x, harmonic_number, coordonnee='y'):
        return 0.0


    # TODO la faire plus jolie ?
    def calculate_magnetic_field(self, Z, Y, X, harmonic_number=1, coordonnee='y'):
        if (type(Z) == np.ndarray):
            B = np.zeros_like(Z)
        else:
            if (type(Y) == np.ndarray):
                B = np.zeros_like(Y)
            else:
                if (type(X) == np.ndarray):
                    B = np.zeros_like(X)
                else:
                    B = self.fct_magnetic_field(z=Z, y=Y, x=X,
                                                harmonic_number=harmonic_number, coordonnee=coordonnee)

        if (type(Z) == np.ndarray):
            if (type(Y) == np.ndarray):
                if (len(Y) != len(Z)):
                    raise Exception(' Y and Z must have the same lenght')
                if (type(X) == np.ndarray):
                    if (len(X) != len(Z)):
                        raise Exception(' X and Z must have the same lenght')
                    for i, Zi in enumerate(Z):
                        B[i] = self.fct_magnetic_field(z=Zi, y=Y[i], x=X[i],
                                                   harmonic_number=harmonic_number, coordonnee=coordonnee)
                else:
                    for i, Zi in enumerate(Z):
                        B[i] = self.fct_magnetic_field(z=Zi, y=Y[i], x=X,
                                                       harmonic_number=harmonic_number, coordonnee=coordonnee)
            else:
                if (type(X) == np.ndarray):
                    if (len(X) != len(Z)):
                        raise Exception(' X and Z must have the same lenght')
                    for i, Zi in enumerate(Z):
                        B[i] = self.fct_magnetic_field(z=Zi, y=Y, x=X[i],
                                                       harmonic_number=harmonic_number, coordonnee=coordonnee)
                else:
                    for i, Zi in enumerate(Z):
                        B[i] = self.fct_magnetic_field(z=Zi, y=Y, x=X,
                                                       harmonic_number=harmonic_number, coordonnee=coordonnee)
        else:
            if (type(Y) == np.ndarray):
                if (type(X) == np.ndarray):
                    if (len(X) != len(Y)):
                        raise Exception(' X and Z must have the same lenght')
                for i, Yi in enumerate(Y):
                    B[i] = self.fct_magnetic_field(z=Z, y=Yi, x=X[i], harmonic_number=harmonic_number,
                                                   coordonnee=coordonnee)
            else:
                if (type(X) == np.ndarray):
                    for i, Xi in enumerate(X):
                        B[i] = self.fct_magnetic_field(z=Z, y=Y, x=Xi, harmonic_number=harmonic_number,
                                                       coordonnee=coordonnee)
                else:
                    B = self.fct_magnetic_field(z=Z, y=Y, x=X, harmonic_number=harmonic_number, coordonnee=coordonnee)
        return B


    # a Magnetic structur create a magnetic field
    # the object MagneticField is create like :
    # Bx , By , Bz are function R**(3) -> R
    # this function depend of the magnet type (BendingMagnet or undulator ..)
    # X, Y can be array or number, they describe the area where we want to work
    # Z must be an array , it will be use later
    # they have not necessary the same len
    # peux etre a changer et ne metrre que les fonctions ... oui
    def create_magnetic_field(self, harmonic_number=1):
        By = (lambda z, y, x: self.calculate_magnetic_field(Z=z, Y=y, X=x, harmonic_number=harmonic_number, coordonnee='y'))
        Bz = (lambda z, y, x: self.calculate_magnetic_field(Z=z, Y=y, X=x, harmonic_number=harmonic_number, coordonnee='z'))
        Bx = (lambda z, y, x: self.calculate_magnetic_field(Z=z, Y=y, X=x, harmonic_number=harmonic_number, coordonnee='x'))
        B = MagneticField(Bx, By, Bz)
        return B



if __name__ == "__main__" :
    pass



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"


import numpy as np
class ElectricalField(object):
   def __init__(self, electrical_field, X, Y, distance):
      # TODO: maybe check constants if we just calcualte vector potential A
      self._electrical_field = electrical_field.copy()
      self._X = X.copy()
      self._Y = Y.copy()
      self._distance = distance

   def X(self):
      return self._X.copy()

   def Y(self):
      return self._Y.copy()

   def distance(self):
      return self._distance

   def electrical_field(self):
      return self._electrical_field.copy()



   def intensity(self):
         return np.sum(np.abs(self.electrical_field()) ** 2, axis=-1)







# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np


class Trajectory(object):

    def __init__(self, t, x, y, z, v_x, v_y, v_z, a_x, a_y, a_z):
        self.t = t
        self.x = x
        self.y = y
        self.z = z
        self.v_x = v_x
        self.v_y = v_y
        self.v_z = v_z
        self.a_x = a_x
        self.a_y = a_y
        self.a_z = a_z


    def nb_points(self):
        N = len(self.t)
        if (len(self.x) == N and len(self.y) == N
            and len(self.z) == N and len(self.v_x) == N
            and len(self.v_y) == N and len(self.v_z) == N
            and len(self.a_x) == N and len(self.a_y) == N
            and len(self.a_z) == N):
            N = len(self.t)
        else: # raise Exception("different lenght in trajectory.")
            print("different lenght in trajectory.")
            N = 0
        return N

    #TODO: compute By versus z (proportional to acceleration)
    # def Bz_versus_z(self):
    #     cte = 1.0
    #     return self.a_y * cte

    def convert(self, T):
        if T.shape[0] == 10:
            self.t = T[0]
            self.x = T[1]
            self.y = T[2]
            self.z = T[3]
            self.v_x = T[4]
            self.v_y = T[5]
            self.v_z = T[6]
            self.a_x = T[7]
            self.a_y = T[8]
            self.a_z = T[9]


    def multiply_by(self, cst):
        self.x *= cst
        self.y *= cst
        self.z *= cst
        self.v_x *= cst
        self.v_y *= cst
        self.v_z *= cst
        self.a_x *= cst
        self.a_y *= cst
        self.a_z *= cst

    def average_vz(self):
        Beta_et_calculated=np.sum(self.v_z) / len(self.v_z)
        return Beta_et_calculated

    def difference_with(self, trajectory_test):
        #TODO remetre ?
        #if any(np.abs(self.t - trajectory_test.t) > np.abs(self.t).max()*1e-6):
        #    raise Exception("Problem : the two trajectory have not the same vector t ??")
        trajec_test=trajectory_test
        error = np.zeros((10, self.nb_points()))
        error[0] = self.t.copy()
        error[1] = np.abs(self.x - trajec_test.x)
        error[2] = np.abs(self.y - trajec_test.y)
        error[3] = np.abs(self.z - trajec_test.z)
        error[4] = np.abs(self.v_x - trajec_test.v_x)
        error[5] = np.abs(self.v_y - trajec_test.v_y)
        error[6] = np.abs(self.v_z - trajec_test.v_z)
        error[7] = np.abs(self.a_x - trajec_test.a_x)
        error[8] = np.abs(self.a_y - trajec_test.a_y)
        error[9] = np.abs(self.a_z - trajec_test.a_z)
        return Trajectory(error[0],error[1],error[2],error[3],error[4],error[5],error[6]
                           ,error[7],error[8],error[9])


    def error_max(self, trajectory_test):
        if any((self.t - trajectory_test.t) > np.abs(self.t).max() * 1e-6):
            raise Exception("Problem : the two trajectory have not the same vector t ??")
        error = np.zeros(10)
        error[0] = self.nb_points()
        error[1] = (np.abs(self.x - trajectory_test.x)).max()
        error[2] = (np.abs(self.y - trajectory_test.y)).max()
        error[3] = (np.abs(self.z - trajectory_test.z)).max()
        error[4] = (np.abs(self.v_x - trajectory_test.v_x)).max()
        error[5] = (np.abs(self.v_y - trajectory_test.v_y)).max()
        error[6] = (np.abs(self.v_z - trajectory_test.v_z)).max()
        error[7] = (np.abs(self.a_x - trajectory_test.a_x)).max()
        error[8] = (np.abs(self.a_y - trajectory_test.a_y)).max()
        error[9] = (np.abs(self.a_z - trajectory_test.a_z)).max()
        return error


    def relativ_difference_with(self, trajectory_test):
        #TODO
       # if any((self.t - trajectory_test.t) > np.abs(self.t).max() * 1e-6):
        #    raise Exception("Problem : the two trajectory have not the same vector t ??")
        trajectory_diff = self.difference_with(trajectory_test)
        xm=np.abs(self.x).max()
        if xm != 0.0:
            trajectory_diff.x *= 1.0 / xm

        ym = np.abs(self.y).max()
        if ym != 0.0 :
            trajectory_diff.y *= 1.0 /ym

        zm = np.abs(self.z).max()
        if zm != 0.0 :
            trajectory_diff.z *= 1.0 /zm

        vxm = np.abs(self.v_x).max()
        if vxm != 0.0:
            trajectory_diff.v_x *= 1.0 / vxm

        vym = np.abs(self.v_y).max()
        if vym != 0.0:
            trajectory_diff.v_y *= 1.0 / vym

        vzm = np.abs(self.v_z).max()
        if vzm != 0.0:
            trajectory_diff.v_z *= 1.0 / vzm

        axm = np.abs(self.a_x).max()
        if axm != 0.0:
            trajectory_diff.a_x *= 1.0 / axm

        aym = np.abs(self.a_y).max()
        if aym != 0.0:
            trajectory_diff.a_y *= 1.0 / aym

        azm = np.abs(self.a_z).max()
        if azm != 0.0:
            trajectory_diff.a_z *= 1.0 / azm

        return trajectory_diff



    def plot_3D(self,title="",label='trajectory'):
        import matplotlib.pyplot as plt
        import scipy.constants as codata

        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = fig.gca(projection='3d')
        ax.plot(self.z*codata.c, self.x*codata.c, self.y*codata.c, label=label)
        ax.legend()
        #ax.autoscale(enable=True, axis='x', tight=None)
        plt.title(title)
        ax.set_xlabel("Z [m]")
        ax.set_ylabel('X [m]')
        ax.set_zlabel("Y [m]")
        plt.show()

    def plot_2D(self,title=""):
        import matplotlib.pyplot as plt
        import scipy.constants as codata

        plt.plot(self.z*codata.c, self.x*codata.c)
        plt.title(title)
        plt.xlabel("Z [m]")
        plt.ylabel('X [m]')
        plt.show()

    def plot(self,title=""):
        import matplotlib.pyplot as plt


        plt.plot(self.t, self.x)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('X')
        plt.show()

        plt.plot(self.t, self.y)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Y')
        plt.show()

        plt.plot(self.t, self.z)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Z')
        plt.show()

        print('average of Vz  =')
        Beta_et = np.sum(self.v_z) / len(self.v_z)
        print(Beta_et)

        Z = Beta_et * self.t
        plt.plot(self.t, self.z - Z)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Z - Beta* t')
        plt.show()

        plt.plot(self.t, self.v_x)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Vx')
        plt.show()

        plt.plot(self.t, self.v_y)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Vy')
        plt.show()

        plt.plot(self.t, self.v_z)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Vz')
        plt.show()

        plt.plot(self.t, self.v_z - Beta_et)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Vz - Beta*')
        plt.show()

        plt.plot(self.t, self.a_x)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Ax')
        plt.show()

        plt.plot(self.t, self.a_y)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Ay')
        plt.show()

        plt.plot(self.t, self.a_z)
        plt.title(title)
        plt.xlabel('t')
        plt.ylabel('Az')
        plt.show()


    # plot all coordinate of the 2 trajectories in function of the time
    # It must have the same shape and the same fistr vector wich represent time.
    def plot_2_trajectory(self, traj2):
        import matplotlib.pyplot as plt
        plt.plot(self.t, self.x)
        plt.plot(traj2.t, traj2.x)
        plt.title(" X = f(t) ")
        plt.xlabel('t')
        plt.ylabel('X')
        plt.show()

        plt.plot(self.t, self.y)
        plt.plot(traj2.t, traj2.y)
        plt.title(" Y = f(t) ")
        plt.xlabel('t')
        plt.ylabel('V')
        plt.show()

        plt.plot(self.t, self.z)
        plt.plot(traj2.t, traj2.z)
        plt.title(" Z  = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Z')
        plt.show()

        Beta_et = np.sum(self.v_z) / len(self.v_z)
        plt.plot(self.t, self.z - Beta_et * self.t)
        plt.plot(traj2.t, traj2.z - Beta_et * self.t)
        plt.title(" Z -Beta* t = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Z')
        plt.show()

        plt.plot(self.t, self.v_x)
        plt.plot(traj2.t, traj2.v_x)
        plt.title(" Vx = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Vx')
        plt.show()

        plt.plot(self.t, self.v_y)
        plt.plot(traj2.t, traj2.v_y)
        plt.title(" Vy = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Vy')
        plt.show()

        plt.plot(self.t, self.v_z)
        plt.plot(traj2.t, traj2.v_z)
        plt.title(" Vz = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Vz')
        plt.show()

        plt.plot(self.t, self.a_x)
        plt.plot(traj2.t, traj2.a_x)
        plt.title(" Ax = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Ax')
        plt.show()

        plt.plot(self.t, self.a_y)
        plt.plot(traj2.t, traj2.a_y)
        plt.title(" Ay = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Ay')
        plt.show()

        plt.plot(self.t, self.a_z)
        plt.plot(traj2.t, traj2.a_z)
        plt.title(" Az = f(t) ")
        plt.xlabel('t')
        plt.ylabel('Az')
        plt.show()

    def copy(self):
        return Trajectory(t=self.t.copy(), x=self.x.copy(), y=self.y.copy(), z=self.z.copy(), v_x=self.v_x.copy(),
                          v_y=self.v_y.copy(), v_z=self.v_z.copy(), a_x=self.a_x.copy(),
                          a_y=self.a_y.copy(), a_z=self.a_z.copy())







if __name__ == "__main__" :
    t=np.linspace(0.0,20.0,1001)
    x = np.sin(t)
    y= 0.0*t
    z=np.cos(t)
    v_x=np.cos(t)
    v_y=0.0*t
    v_z=-np.sin(t)
    a_x=-np.sin(t)
    a_y=0.0*t
    a_z=-np.cos(t)
    Traj=Trajectory(t,x,y,z,v_x,v_y,v_z,a_x,a_y,a_z)
    Traj.plot()
    Traj.plot_3D()

    Traj2=Traj.copy()
    Traj2.multiply_by(3)

    Traj.plot_2_trajectory(Traj2)

    Diff=Traj.difference_with(Traj2)
    Diff.plot()
    Diff.plot_3D()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import unittest
import numpy as np
from pySRU.Trajectory import Trajectory


class TrajectoryTest(unittest.TestCase):
    def test_error_nb_pts_copy(self):
        t = np.arange(0.0, 6.0*np.pi, 0.005*np.pi)
        x = np.sin(t)
        y = 0.0 * t
        z = np.cos(t)
        v_x = np.cos(t)
        v_y = 0.0 * t
        v_z = -np.sin(t)
        a_x = -np.sin(t)
        a_y = 0.0 * t
        a_z = -np.cos(t)
        Traj = Trajectory(t, x, y, z, v_x, v_y, v_z, a_x, a_y, a_z)

        Traj2 = Traj.copy()

        self.assertEqual(Traj.nb_points(), 1200)


        self.assertTrue(all(Traj.t==Traj2.t))

        error=Traj.error_max(Traj2)
        # Traj.draw_2_trajectory(Traj2)
        error[0] -= len(Traj.t)
        self.assertEqual(error.max(), 0.0)

        Traj2.multiply_by(3)
        error = Traj.error_max(Traj2)
        error[0] -= len(Traj.t)
        self.assertEqual(error.max(), 2.0)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
import scipy.constants as codata
import scipy.special as special
from pySRU.MagneticStructure import MagneticStructure, PLANE_UNDULATOR,BENDING_MAGNET


#TODO changer et mettre la divergeance en parametre ?
class MagneticStructureBendingMagnet(MagneticStructure):
    def __init__(self, Bo, horizontale_divergeance,electron_energy=1e9):
        super(self.__class__, self).__init__(magnet_type=BENDING_MAGNET)
        self.horizontal_div=horizontale_divergeance
        self.radius_curv=electron_energy*1e9/(codata.c*Bo)
        self.length=np.sin(horizontale_divergeance)*self.radius_curv
        self.Bo = Bo


    def copy(self):
        return MagneticStructureBendingMagnet(length=self.length, Bo=self.Bo)

    def fct_magnetic_field(self, z, y, x, harmonic_number, coordonnee='y'):
        Zo = self.length * 0.6
        if coordonnee == 'y' and z >= -Zo and z <= Zo:
            # codata.m_e * codata.c / codata.e= 0.00170450894933
            Bo = self.Bo

        else:  # coordonnee == 'z' :
            Bo = 0.0
        return Bo


    def fct_magnetic_field2(self, z, y, x, harmonic_number, coordonnee='y'):
        Zo = self.length * 0.5
        if coordonnee == 'y':
            # codata.m_e * codata.c / codata.e= 0.00170450894933
            Bo = self.Bo
        else:  # coordonnee == 'z' :
            Bo = 0.0

            # we enelarge the real effect of the magnetic field by 4 lambda_u
            # on each extremity of the undulator
        L_magn_field = self.length * 1.5

        # we know considere that if -(L/2 + lambda_u/4) < Z < (L/2 + lambda_u/4)
        # the magnetic field if a classic cosinus (or sinus)
        L_cosinus_part = -Zo

        if ((z < -L_magn_field) or (z > L_magn_field)):
            dB = 0.0

        else:
            if (z < -L_cosinus_part or z > L_cosinus_part):

                # in this case, we work with a gaussian,
                # so we shift the coordinate frame for use a central gaussian
                if z > 0.0:
                    z_shift = z - L_cosinus_part
                else:
                    z_shift = z + L_cosinus_part

                sigma = (L_magn_field - L_cosinus_part) / 5.0
                # dB= Bo*(1.0-z_shift*self.get_L()**2/2.0)*np.exp(-0.5*(z_shift/sigma)**2)
                dB = Bo * np.exp(-0.5 * (z_shift / sigma) ** 2)
            else:
                # in this case we work in the cosinus part
                dB = Bo
        return dB



    def print_parameters(self):
        print(" Magnetic Structure (Bending Magnet):")
        print('    length : %.5f (m)'%self.length)
        print('    magnetic_field_strength : %.5f'%self.Bo)


if __name__ == "__main__" :

    BM = MagneticStructureBendingMagnet(Bo=0.85, length=1e-3 * 20.0)

    print('bending magnet test :')
    BM.print_parameters()
    B = BM.create_magnetic_field()
    Z=np.linspace(-BM.length/2.0,BM.length/2.0,10000)
    B.plot_z(Z=Z,Y=0.0,X=0.0)




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import unittest
import numpy as np
from scipy import integrate
from pySRU.Radiation import Radiation


class RadiationTest(unittest.TestCase):

    def test_first(self):

        X=np.linspace(0.0,0.5,101)
        Y=np.linspace(0.0,0.5,101)
        distance=100
        X_grid,Y_grid=np.meshgrid(X,Y)
        intensity_gird = X_grid ** 2 * Y_grid ** 2
        intensity_list= X ** 2 * Y ** 2
        radGrid=Radiation(intensity=intensity_gird, X=X_grid, Y=Y_grid,distance=distance)
        radList=Radiation(intensity=intensity_list, X=X, Y=Y,distance=distance)

        self.assertEqual(radGrid.max(), radList.max())


        # test RadiationGrid

        # test copy
        radGrid2=radGrid.copy()
        self.assertTrue(radGrid.XY_are_similar_to(radGrid2))

        err=radGrid.error_max(radGrid2)
        self.assertEqual(err.max(),0.0)

        # test modification of the screen
        radGrid2.X += 1
        self.assertFalse(radGrid.XY_are_similar_to(radGrid2))

        radGrid3=radGrid.copy()
        radGrid3.change_Nb_pts(Nb_pts=120)
        self.assertFalse(radGrid.XY_are_similar_to(radGrid3))
        self.assertTrue(len(radGrid3.X.shape)==2)
        self.assertTrue(len(radGrid3.Y.shape) == 2)
        self.assertTrue(radGrid.X.max()==radGrid3.X.max())
        self.assertTrue(radGrid.X.min() == radGrid3.X.min())
        self.assertTrue(radGrid.Y.max() == radGrid3.Y.max())
        self.assertTrue(radGrid.Y.min() == radGrid3.Y.min())

        #error
        radGrid2.intensity= radGrid2.X ** 2 * radGrid2.Y ** 2

        err=radGrid.relativ_error(radGrid2)
        self.assertEqual(err,8.0)




        # test RadiationList

        # test copy
        radList2=radList.copy()
        self.assertTrue(radList.XY_are_similar_to(radList2))

        err = radList.error_max(radList2)
        self.assertEqual(err.max(), 0.0)

        # test modification of the screen
        radList2.X += 1
        self.assertFalse(radList.XY_are_similar_to(radList2))

        radList3=radList.copy()
        radList3.change_Nb_pts(Nb_pts=120)
        self.assertFalse(radList.XY_are_similar_to(radList3))
        self.assertTrue(len(radList3.X.shape)==1)
        self.assertTrue(len(radList3.Y.shape) == 1 )
        self.assertTrue(radList.X.max()==radList3.X.max())
        self.assertTrue(radList.X.min() == radList3.X.min())
        self.assertTrue(radList.Y.max() == radList3.Y.max())
        self.assertTrue(radList.Y.min() == radList3.Y.min())


        #test error
        radList2.intensity = radList2.X ** 2 * radList2.Y ** 2

        err=radGrid.relativ_error(radGrid2)
        self.assertEqual(err,8.0)



    def test_integration(self):
        X = np.linspace(0.0, 1., 101)
        Y = np.linspace(0.0, 1., 101)
        distance = 100
        X_grid, Y_grid = np.meshgrid(X, Y)
        intensity_gird = X_grid + Y_grid
        X *= 0.0
        intensity_list = X + Y
        radGrid = Radiation(intensity=intensity_gird, X=X_grid, Y=Y_grid, distance=distance)
        radList = Radiation(intensity=intensity_list, X=X, Y=Y, distance=distance)

        integartionGrid=radGrid.integration(use_flux_per_mrad2_or_mm2=0)
        integrationList=radList.integration(use_flux_per_mrad2_or_mm2=0)
        self.assertEqual(integrationList,0.5)
        self.assertEqual(integartionGrid,0.5+integrationList)




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
import scipy.constants as codata
from abc import abstractmethod
from pySRU.MagneticField import MagneticField
from pySRU.ElectronBeam import ElectronBeam
from pySRU.MagneticStructure import MagneticStructure,PLANE_UNDULATOR,BENDING_MAGNET




# magnetic field est il interressant ici ???
class Source(object):
    def __init__(self,electron_beam,magnetic_field,magnetic_structure):
        self.electron_beam=electron_beam
        self.magnetic_structure=magnetic_structure
        if magnetic_field==None :
            self.magnetic_field=self.magnetic_structure.create_magnetic_field()
        else :
            self.magnetic_field=magnetic_field


    @abstractmethod
    def copy(self):
        return

    # gueter
    def I_current(self):
        return self.electron_beam.I_current

    def Electron_energy(self):
        return self.electron_beam.Electron_energy

    # utiles
    def Lorentz_factor(self):
        return self.electron_beam.Lorentz_factor()

    #utile ?
    def electron_speed(self):#BETA
        gamma = self.electron_beam.Lorentz_factor()
        Beta = np.sqrt(1.0 - 1.0 / gamma ** 2)
        return Beta

    def magnet_type(self):
        return self.magnetic_structure.magnet_type


    @abstractmethod
    def magnetic_field_strength(self):
        return

    def atol_for_ODE_method(self):
        return 1e-11

    def rtol_for_ODE_method(self):
        return 1e-11

    # constante de construction

    @abstractmethod
    def choose_distance_automatic(self, alpha=2,photon_frequency=None):
        return

    @abstractmethod
    def choose_nb_pts_trajectory(self, alpha,photon_frequency):
        return

    @abstractmethod
    def choose_initial_contidion_automatic(self):
        return

    @abstractmethod
    def choose_photon_frequency(self):
        return

    @abstractmethod
    def angle_deflection_max(self):
        return


    @abstractmethod
    def analytical_times_vector(self,Nb_pts):
        return

    @abstractmethod
    def construct_times_vector(self,initial_contition,Nb_pts):
        return


    # source parameter


    def print_parameters(self) :
        self.electron_beam.print_parameters()
        self.magnetic_structure.print_parameters()

    # theory
    @abstractmethod
    def theoretical_flux_on_axis(self,frequency):
        return

    @abstractmethod
    def theoretical_flux_on_axis_omega_scan(self, n, omega_array) :
        return







if __name__ == "__main__" :

    from SourceUndulatorPlane import SourceUndulatorPlane
    from MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane
    electron_beam_test=ElectronBeam(Electron_energy=1.3, I_current=1.0)
    undulator_test=MagneticStructureUndulatorPlane(K=1.87,period_length=0.035,length=0.49)
    source_test = SourceUndulatorPlane(undulator=undulator_test, electron_beam=electron_beam_test,
                                       magnetic_field=None)
    print(type(source_test.magnetic_field))
    print(type(source_test.magnetic_field.Bx))
    print(type(source_test.magnetic_field.By))
    print(source_test.magnetic_field.Bx(0.0, 0.0, 0.0))

    #Exemple1_undulator()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
from pySRU.Radiation import Radiation
import scipy.constants as codata
import scipy.integrate as integrate
from pySRU.TrajectoryFactory import TrajectoryFactory, TRAJECTORY_METHOD_ANALYTIC , TRAJECTORY_METHOD_ODE
from pySRU.ElectricalField import ElectricalField
from scipy import constants as codata



RADIATION_METHOD_NEAR_FIELD = 0        # calls self.energy_radiated_near_field2
RADIATION_METHOD_APPROX = 1            # calls self.energy_radiated_approx2
RADIATION_METHOD_APPROX_FARFIELD = 2   # calls self.energy_radiated_approximation_and_farfield2
# calls self.energy_radiated_near_field
# calls self.energy_radiated_approx
# calls self.energy_radiated_approximation_and_farfield
eV_to_J = codata.e

class RadiationFactory(object):
    def __init__(self, method, photon_frequency):
        self.photon_frequency=photon_frequency
        self.method=method


    def copy(self):
        return RadiationFactory(method=self.method, photon_frequency=self.photon_frequency)

    def energy_eV(self):
        return self.photon_frequency * codata.hbar / eV_to_J

    # Photon's flow all over a screen situate at distance D of an undulator
    def create_for_one_relativistic_electron(self, trajectory, source, XY_are_list=False, Nb_pts=101, distance=None, X=None, Y=None):
        if X is None or Y is None:
            print('calculate X and Y array')
            if distance == None:
                if self.method==RADIATION_METHOD_APPROX_FARFIELD :#TODO rendre distance = None possible por nearfield ?
                    xy_max = np.tan(1./ source.Lorentz_factor())
                else :
                    distance=source.choose_distance_automatic(2)
                    xy_max = np.tan(1. / source.Lorentz_factor())*distance
            else :
                xy_max = np.tan(1. / source.Lorentz_factor())*distance
            X = np.linspace(0.0, xy_max, Nb_pts)
            Y = np.linspace(0.0, xy_max, Nb_pts)
        else :
            if type(X) != np.ndarray:
                X = np.linspace(0.0, X, Nb_pts)
            if type(Y) != np.ndarray:
                Y = np.linspace(0.0, Y, Nb_pts)



        if not XY_are_list :
            # Y, X = np.meshgrid(X, Y)
            X_ones = np.ones_like(X)
            Y_ones = np.ones_like(Y)
            X = np.outer(X,Y_ones)
            Y = np.outer(X_ones,Y)

        intensity = self.calculate_radiation_intensity(trajectory=trajectory, source=source,
                                                       distance=distance, X_array=X, Y_array=Y)

        radiation = Radiation(intensity=intensity, X=X, Y=Y, distance=distance)

        return radiation

    def calculate_electrical_field(self, trajectory, source, X_array, Y_array, distance=None):
        # c1 = codata.e ** 2 * omega1 ** 2 / (16 * np.pi ** 3 * codata.epsilon_0 * codata.c )
        # c2 = I / codata.e  # multiply by number of electrons in 1 A
        # c3 = 2.0*np.pi / (codata.h * omega1)  # divide by e energy (to get number of photons)
        # c4 = 1e-3 / omega1  # to get .1% energy bandwidth  #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        # c5 = 1e-3 * 1e-3  # from rad to .1mrad angle bandwidth



        c6 = codata.e * source.I_current() * 1e-9 / (8.0 * np.pi ** 2 * codata.epsilon_0 * codata.c * codata.h)

        # to guarantee the flux in phot/mm2 reduces with 1/r2
        if distance != None:
            c6 /= distance**2

        if X_array.size != Y_array.size:
            raise Exception("X and Y dimensions must be equal.")

        shape1 = X_array.shape
        X = X_array.flatten()
        Y = Y_array.flatten()

        res = np.zeros((X.size, 3), dtype=np.complex128)

        gamma = source.Lorentz_factor()


        #
        # from more exact to less exact/precise the ranging is:
        #   energy_radiated_near_field2   <-- RADIATION_METHOD_NEAR_FIELD
        #   energy_radiated_approx2       <-- RADIATION_METHOD_APPROX
        #   energy_radiated_near_field
        #   energy_radiated_approx
        #   energy_radiated_approximation_and_farfield  <--  RADIATION_METHOD_APPROX_FARFIELD
        #   energy_radiated_approximation_and_farfield


        if self.method == RADIATION_METHOD_NEAR_FIELD:  # 0
            calculation_function = self.energy_radiated_near_field2
        elif self.method == RADIATION_METHOD_APPROX: # 1
            calculation_function = self.energy_radiated_approx2
        elif RADIATION_METHOD_APPROX_FARFIELD:
            calculation_function = self.energy_radiated_approximation_and_farfield
        else:
            raise Exception("Method code not implemented.")


        # TODO: Possible missing imaginary phase in constant?
        for i in range(len(X)):
            res[i, :] = calculation_function(trajectory=trajectory, distance=distance,
                                             gamma=gamma, x=X[i], y=Y[i])

        res *= c6**0.5

        # if len(shape1)==2 :
        #     res = res.reshape((shape1[0], shape1[1], 3))
        # else :
        #     res = res.reshape(shape1, 3)
        res = res.reshape(shape1 + (3,))

        electrical_field = ElectricalField(electrical_field=res, X=X, Y=Y, distance=distance)

        return electrical_field

    # Photon's flow all over a screen situate at distance D of an undulator
    def calculate_radiation_intensity(self, trajectory, source, X_array, Y_array, distance=None):
        electrical_field = self.calculate_electrical_field(trajectory, source, X_array, Y_array, distance)
        return electrical_field.intensity()

    def energy_radiated_approximation_and_farfield2(self,trajectory, gamma, x, y, distance):

        # N = trajectory.shape[1]
        N = trajectory.nb_points()
        if distance == None:
            # in radian :
            n_chap = np.array([x, y, 1.0 - 0.5 * (x ** 2 + y ** 2)])
            X = np.sqrt(x ** 2 + y ** 2 )#TODO a changer
        #in meters :
        else :
            X = np.sqrt(x ** 2 + y ** 2 + distance ** 2)
            n_chap = np.array([x, y, distance]) / X

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)
        A1 = (n_chap[0] * trajectory.a_x + n_chap[1] * trajectory.a_y + n_chap[2] * trajectory.a_z)
        A2 = (n_chap[0] * (n_chap[0] - trajectory.v_x) + n_chap[1] * (n_chap[1] - trajectory.v_y)
                + n_chap[2] * (n_chap[2] - trajectory.v_z))
        Alpha2 = np.exp(
            0. + 1j * self.photon_frequency * (trajectory.t + X / codata.c - n_chap[0] * trajectory.x
                                               - n_chap[1] * trajectory.y - n_chap[2] * trajectory.z))

        Alpha1 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x
                             - n_chap[1] * trajectory.v_y - n_chap[2] * trajectory.v_z)) ** 2

        integrand[0] += (A1 * (n_chap[0] - trajectory.v_x) - A2 * trajectory.a_x) * Alpha2 * Alpha1
        integrand[1] += (A1 * (n_chap[1] - trajectory.v_y) - A2 * trajectory.a_y) * Alpha2 * Alpha1
        integrand[2] += (A1 * (n_chap[2] - trajectory.v_z) - A2 * trajectory.a_z) * Alpha2 * Alpha1
        for k in range(3):
            # E[k] = np.trapz(integrand[k], self.trajectory.t)
            E[k] = np.trapz(integrand[k], trajectory.t)

        return E

    def energy_radiated_approximation_and_farfield(self, trajectory, gamma, x, y, distance):
        # N = trajectory.shape[1]
        N = trajectory.nb_points()
        if distance == None:
            # in radian :
            n_chap = np.array([x, y, 1.0 - 0.5 * (x ** 2 + y ** 2)])
            X = np.sqrt(x ** 2 + y ** 2 )#TODO a changer
        #in meters :
        else :
            X = np.sqrt(x ** 2 + y ** 2 + distance ** 2)
            n_chap = np.array([x, y, distance]) / X

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)
        A1 = (n_chap[1] * trajectory.v_z - n_chap[2] * trajectory.v_y)
        A2 = (-n_chap[0] * trajectory.v_z + n_chap[2] * trajectory.v_x)
        A3 = (n_chap[0] * trajectory.v_y - n_chap[1] * trajectory.v_x)
        Alpha2 = np.exp(
            0. + 1j * self.photon_frequency * (trajectory.t + X / codata.c - n_chap[0] * trajectory.x
                                               - n_chap[1] * trajectory.y - n_chap[2] * trajectory.z))


        integrand[0] -= ( n_chap[1]*A3 - n_chap[2]*A2) * Alpha2
        integrand[1] -= (- n_chap[0]*A3 + n_chap[2]*A1) * Alpha2
        integrand[2] -= ( n_chap[0]*A2 - n_chap[1]*A1) * Alpha2

        for k in range(3):
            # E[k] = np.trapz(integrand[k], self.trajectory.t)
            E[k] = np.trapz(integrand[k], trajectory.t)
        E *= self.photon_frequency * 1j

        terme_bord = np.full((3), 0. + 1j * 0., dtype=complex)
        Alpha_1 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x[-1]
                          - n_chap[1] * trajectory.v_y[-1] - n_chap[2] * trajectory.v_z[-1]))
        Alpha_0 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x[0]
                          - n_chap[1] * trajectory.v_y[0] - n_chap[2] * trajectory.v_z[0]))

        terme_bord += ((n_chap[1] * A3[-1] - n_chap[2] * A2[-1]) * Alpha_1 *
                       Alpha2[-1])
        terme_bord -= ((n_chap[1] * A3[0] - n_chap[2] * A2[0]) * Alpha_0 *
                       Alpha2[0])
        E += terme_bord

        return E



    def energy_radiated_farfield2(self, trajectory, gamma, x, y, distance):
        N = trajectory.nb_points()
        if distance == None:
            # in radian :
            n_chap = np.array([x, y, 1.0 - 0.5 * (x ** 2 + y ** 2)])
            X = np.sqrt(x ** 2 + y ** 2 )#TODO a changer
        #in meters :
        else :
            X = np.sqrt(x ** 2 + y ** 2 + distance ** 2)
            n_chap = np.array([x, y, distance]) / X

        R= X/codata.c - n_chap[0] * trajectory.x- n_chap[1] * trajectory.y - n_chap[2] * trajectory.z

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)
        A1 = (n_chap[0] * trajectory.a_x + n_chap[1] * trajectory.a_y + n_chap[2] * trajectory.a_z)
        A2 = (n_chap[0] * (n_chap[0] - trajectory.v_x) + n_chap[1] * (n_chap[1] - trajectory.v_y)
                + n_chap[2] * (n_chap[2] - trajectory.v_z))
        Alpha2 = np.exp(0. + 1j * self.photon_frequency * (trajectory.t + R))
        Alpha1 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x
                             - n_chap[1] * trajectory.v_y - n_chap[2] * trajectory.v_z)) ** 2
        cst=codata.c/(R*gamma**2)
        integrand[0] += (A1 * (n_chap[0] - trajectory.v_x) - A2 * trajectory.a_x
                         + cst * (n_chap[0] - trajectory.v_x) ) * Alpha2 * Alpha1
        integrand[1] += (A1 * (n_chap[1] - trajectory.v_y) - A2 * trajectory.a_y
                         + cst * (n_chap[1] - trajectory.v_y)  ) * Alpha2 * Alpha1
        integrand[2] += (A1 * (n_chap[2] - trajectory.v_z) - A2 * trajectory.a_z
                         + cst * (n_chap[2] - trajectory.v_z)  ) * Alpha2 * Alpha1
        for k in range(3):
                # E[k] = np.trapz(integrand[k], self.trajectory.t)
            E[k] = np.trapz(integrand[k], trajectory.t)

        E *= -1j * np.exp(1j * self.photon_frequency / codata.c * (n_chap[0] * x + n_chap[1] * y + n_chap[2] * distance))

        return E

    def energy_radiated_farfield(self, trajectory, gamma, x, y, distance):
        # N = trajectory.shape[1]
        N = trajectory.nb_points()
        if distance == None:
            # in radian :
            n_chap = np.array([x, y, 1.0 - 0.5 * (x ** 2 + y ** 2)])
            X = np.sqrt(x ** 2 + y ** 2 )#TODO a changer
        #in meters :
        else :
            X = np.sqrt(x ** 2 + y ** 2 + distance ** 2)
            n_chap = np.array([x, y, distance]) / X

        R= X/codata.c - n_chap[0] * trajectory.x- n_chap[1] * trajectory.y - n_chap[2] * trajectory.z

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)
        A1 = (n_chap[1] * trajectory.v_z - n_chap[2] * trajectory.v_y)
        A2 = (-n_chap[0] * trajectory.v_z + n_chap[2] * trajectory.v_x)
        A3 = (n_chap[0] * trajectory.v_y - n_chap[1] * trajectory.v_x)
        Alpha1 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x
                         - n_chap[1] * trajectory.v_y - n_chap[2] * trajectory.v_z)) ** 2
        Alpha2 = np.exp(
            0. + 1j * self.photon_frequency * (trajectory.t + R))
        cst = codata.c / ((gamma ** 2) * R)
        integrand[0] -= ((n_chap[1] * A3 - n_chap[2] * A2) * self.photon_frequency * 1j
                         + cst * (n_chap[0] - trajectory.v_x) * Alpha1) * Alpha2
        integrand[1] -= ((- n_chap[0] * A3 + n_chap[2] * A1) * self.photon_frequency * 1j
                         + cst * (n_chap[1] - trajectory.v_y) * Alpha1) * Alpha2
        integrand[2] -= ((n_chap[0] * A2 - n_chap[1] * A1) * self.photon_frequency * 1j
                         + cst * (n_chap[2] - trajectory.v_z) * Alpha1) * Alpha2

        for k in range(3):
            E[k] = np.trapz(integrand[k], trajectory.t)
            #E[k] = integrate.simps(integrand[k], trajectory.t)

        terme_bord = np.full((3), 0. + 1j * 0., dtype=complex)
        Alpha_1 = (1.0 / (1.0 - n_chap[0][-1] * trajectory.v_x[-1]
                          - n_chap[1][-1] * trajectory.v_y[-1] - n_chap[2][-1] * trajectory.v_z[-1]))
        Alpha_0 = (1.0 / (1.0 - n_chap[0][0] * trajectory.v_x[0]
                          - n_chap[1][0] * trajectory.v_y[0] - n_chap[2][0] * trajectory.v_z[0]))

        terme_bord += ((n_chap[1][-1] * A3[-1] - n_chap[2][-1] * A2[-1]) * Alpha_1 *
                       np.exp(1j * self.photon_frequency * (trajectory.t[-1] + R[-1] / codata.c)))
        terme_bord -= ((n_chap[1][0] * A3[0] - n_chap[2][0] * A2[0]) * Alpha_0 *
                       np.exp(1j * self.photon_frequency * (trajectory.t[0] + R[0] / codata.c)))
        E += terme_bord


        return E

    # exact equation for the energy radiated
    # warning !!!!!!! far field approximation
    def energy_radiated_approx2(self,trajectory, gamma, x, y, distance):
        N = trajectory.nb_points()
        n_chap = np.array([x - trajectory.x * codata.c, y - trajectory.y * codata.c, distance - trajectory.z * codata.c])
        # R = np.zeros(n_chap.shape[1])
        # for i in range(n_chap.shape[1]):
        #     R[i] = np.linalg.norm(n_chap[:, i])
        #     n_chap[:, i] /= R[i]

        R = np.sqrt( n_chap[0]**2 + n_chap[1]**2 + n_chap[2]**2 )
        n_chap[0,:] /= R
        n_chap[1,:] /= R
        n_chap[2,:] /= R

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)
        A1 = (n_chap[0] * trajectory.a_x + n_chap[1] * trajectory.a_y + n_chap[2] * trajectory.a_z)
        A2 = (n_chap[0] * (n_chap[0] - trajectory.v_x) + n_chap[1] * (n_chap[1] - trajectory.v_y)
              + n_chap[2] * (n_chap[2] - trajectory.v_z))
        Alpha2 = np.exp(0. + 1j * self.photon_frequency * (trajectory.t + R / codata.c))
        Alpha1 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x
                         - n_chap[1] * trajectory.v_y - n_chap[2] * trajectory.v_z)) ** 2
        integrand[0] += ((A1 * (n_chap[0] - trajectory.v_x) - A2 * trajectory.a_x)
                         ) * Alpha2 * Alpha1
        integrand[1] += ((A1 * (n_chap[1] - trajectory.v_y) - A2 * trajectory.a_y)
                         ) * Alpha2 * Alpha1
        integrand[2] += ((A1 * (n_chap[2] - trajectory.v_z) - A2 * trajectory.a_z)
                         ) * Alpha2 * Alpha1
        for k in range(3):
            E[k] = np.trapz(integrand[k], trajectory.t)

            #E[k] = integrate.simps(integrand[k], trajectory.t)

        return E

    def energy_radiated_approx(self, trajectory, gamma, x, y, distance):
        N = trajectory.nb_points()
        n_chap = np.array([x - trajectory.x * codata.c, y - trajectory.y * codata.c, distance - trajectory.z * codata.c])
        R = np.sqrt( n_chap[0]**2 + n_chap[1]**2 + n_chap[2]**2 )
        n_chap[0,:] /= R
        n_chap[1,:] /= R
        n_chap[2,:] /= R

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)

        A1 = (n_chap[1] * trajectory.v_z - n_chap[2] * trajectory.v_y)
        A2 = (-n_chap[0] * trajectory.v_z + n_chap[2] * trajectory.v_x)
        A3 = (n_chap[0] * trajectory.v_y - n_chap[1] * trajectory.v_x)
        Alpha2 = np.exp(0. + 1j * self.photon_frequency * (trajectory.t + R / codata.c))

        integrand[0] -= (n_chap[1] * A3 - n_chap[2] * A2) * Alpha2
        integrand[1] -= (- n_chap[0] * A3 + n_chap[2] * A1) * Alpha2
        integrand[2] -= (n_chap[0] * A2 - n_chap[1] * A1) * Alpha2
        for k in range(3):
            E[k] = np.trapz(integrand[k], trajectory.t)
            #E[k] = integrate.simps(integrand[k], trajectory.t)
        E *= self.photon_frequency * 1j

        terme_bord = np.full((3), 0. + 1j * 0., dtype=complex)
        Alpha_1 = (1.0 / (1.0 - n_chap[0][-1] * trajectory.v_x[-1]
                          - n_chap[1][-1] * trajectory.v_y[-1] - n_chap[2][-1] * trajectory.v_z[-1]))
        Alpha_0 = (1.0 / (1.0 - n_chap[0][0] * trajectory.v_x[0]
                          - n_chap[1][0] * trajectory.v_y[0] - n_chap[2][0] * trajectory.v_z[0]))

        terme_bord += ((n_chap[1][-1] * A3[-1] - n_chap[2][-1] * A2[-1]) * Alpha_1 *
                       np.exp(1j * self.photon_frequency * (trajectory.t[-1] + R[-1] / codata.c)))
        terme_bord -= ((n_chap[1][0] * A3[0] - n_chap[2][0] * A2[0]) * Alpha_0 *
                       np.exp(1j * self.photon_frequency * (trajectory.t[0] + R[0] / codata.c)))
        E += terme_bord

        return E

    # energy radiated without the the far filed approxiamation
    def energy_radiated_near_field2(self, trajectory, gamma, x, y, distance):
        N = trajectory.nb_points()
        # equation with denominator
        n_chap = np.array([x - trajectory.x * codata.c, y - trajectory.y * codata.c, distance - trajectory.z * codata.c])
        R = np.sqrt( n_chap[0, :]**2 + n_chap[1, :]**2 + n_chap[2, :]**2 )
        n_chap[:, :] /= R[:]

        E = np.empty((3,), dtype=complex)
        integrand = np.empty((3, N), dtype=complex)

        A1 = (n_chap[0] * trajectory.a_x + n_chap[1] * trajectory.a_y + n_chap[2] * trajectory.a_z)
        A2 = (n_chap[0] * (n_chap[0] - trajectory.v_x) + n_chap[1] * (n_chap[1] - trajectory.v_y)
              + n_chap[2] * (n_chap[2] - trajectory.v_z))
        Alpha2 = np.exp(0. + 1j * self.photon_frequency * (trajectory.t + R / codata.c))

        Alpha1 = (1.0 / (1.0 - n_chap[0] * trajectory.v_x
                         - n_chap[1] * trajectory.v_y - n_chap[2] * trajectory.v_z)) ** 2
        Alpha3 = codata.c / (gamma ** 2 * R)
        integrand[0] = ((A1 * (n_chap[0] - trajectory.v_x) - A2 * trajectory.a_x)
                         + Alpha3 * (n_chap[0] - trajectory.v_x)
                         )
        integrand[1] = ((A1 * (n_chap[1] - trajectory.v_y) - A2 * trajectory.a_y)
                         + Alpha3 * (n_chap[1] - trajectory.v_y)
                         )
        integrand[2] = ((A1 * (n_chap[2] - trajectory.v_z) - A2 * trajectory.a_z)
                         + Alpha3 * (n_chap[2] - trajectory.v_z)
                         )
        integrand[:, :] *= Alpha2 * Alpha1
          

        h = (trajectory.t[-1]-trajectory.t[0]) / (trajectory.t.size-1)
        for k in range(3):
            E[k] = h*( np.sum(integrand[k, :]) - 0.5 * (integrand[k, 0] + integrand[k, -1]) ) 
        return E


    def energy_radiated_near_field(self, trajectory, gamma, x, y, distance):
        # N = trajectory.shape[1]
        N = trajectory.nb_points()
        n_chap = np.array(
            [x - trajectory.x * codata.c, y - trajectory.y * codata.c, distance - trajectory.z * codata.c])
        R = np.sqrt(n_chap[0] ** 2 + n_chap[1] ** 2 + n_chap[2] ** 2)
        n_chap[0, :] /= R
        n_chap[1, :] /= R
        n_chap[2, :] /= R

        E = np.zeros((3,), dtype=complex)
        integrand = np.zeros((3, N), dtype=complex)

        A1 = (n_chap[1] * trajectory.v_z - n_chap[2] * trajectory.v_y)
        A2 = (-n_chap[0] * trajectory.v_z + n_chap[2] * trajectory.v_x)
        A3 = (n_chap[0] * trajectory.v_y - n_chap[1] * trajectory.v_x)
        Alpha1 = np.exp(
            0. + 1j * self.photon_frequency * (trajectory.t + R/codata.c))
        Alpha2 = codata.c / (gamma ** 2 * R)
        integrand[0] -= ((n_chap[1] * A3 - n_chap[2] * A2) * self.photon_frequency* 1j
                         + Alpha2 * (n_chap[0] - trajectory.v_x)
                         ) * Alpha1
        integrand[1] -= ((-n_chap[0] * A3 + n_chap[2] * A1) * self.photon_frequency * 1j
                         + Alpha2 * (n_chap[1] - trajectory.v_y)
                         ) * Alpha1
        integrand[2] -= ((n_chap[0] * A2 - n_chap[1] * A1) * self.photon_frequency * 1j
                         + Alpha2 * (n_chap[2] - trajectory.v_z)
                         ) * Alpha1
        for k in range(3):
            E[k] = np.trapz(integrand[k], trajectory.t)
            #E[k] = integrate.simps(integrand[k], trajectory.t)

        terme_bord = np.full((3), 0. + 1j * 0., dtype=complex)
        Alpha_1 = (1.0 / (1.0 - n_chap[0][-1] * trajectory.v_x[-1]
                         - n_chap[1][-1] * trajectory.v_y[-1] - n_chap[2][-1] * trajectory.v_z[-1]))
        Alpha_0 = (1.0 / (1.0 - n_chap[0][0] * trajectory.v_x[0]
                          - n_chap[1][0] * trajectory.v_y[0] - n_chap[2][0] * trajectory.v_z[0]))

        terme_bord += ((n_chap[1][-1] * A3[-1] - n_chap[2][-1] * A2[-1]) * Alpha_1*
                       np.exp(1j * self.photon_frequency * (trajectory.t[-1] +R[-1]/codata.c) ))
        terme_bord -= ((n_chap[1][0] * A3[0] - n_chap[2][0] * A2[0]) * Alpha_0*
                       np.exp(1j * self.photon_frequency * (trajectory.t[0] + R[0]/codata.c)))
        # E += terme_bord
        return E



    def get_method(self):
        if self.method == RADIATION_METHOD_NEAR_FIELD:
            method = 'Near field calculation'

        elif self.method == RADIATION_METHOD_APPROX_FARFIELD:
            method = '  Far field and approximation calculation '
        else :
            method=' Problem unknow method'
        return method

    def print_parameters(self):
        print('Radiation ')
        print('    method: %s' %self.get_method())
        print('    energy of the emission: %f eV, omega: %f Hz' % (self.energy_eV(),self.photon_frequency))


def Exemple_FARFIELD():

    from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane as Undulator
    from pySRU.ElectronBeam import ElectronBeam
    from pySRU.SourceUndulatorPlane import SourceUndulatorPlane
    from pySRU.TrajectoryFactory import TrajectoryFactory,TRAJECTORY_METHOD_ODE,TRAJECTORY_METHOD_ANALYTIC
    import time



    undulator_test = Undulator(K=1.87, period_length=0.035, length=0.035 * 14)
    electron_beam_test = ElectronBeam(Electron_energy=1.3, I_current=1.0)
    magnetic_field_test = undulator_test.create_magnetic_field()


    magnetic_field_test.plot_z(0,0,np.linspace(-0.035 * (14+8) / 2,0.035 * (14+8) / 2,500))


    source_test = SourceUndulatorPlane(undulator=undulator_test,
                         electron_beam=electron_beam_test,
                         magnetic_field=magnetic_field_test)

    traj = TrajectoryFactory(Nb_pts=2000, method=TRAJECTORY_METHOD_ODE).create_from_source(source_test)

    t0 = time.time()
    Rad = RadiationFactory(photon_frequency=source_test.harmonic_frequency(1), method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts=101
                           ).create_for_one_relativistic_electron(trajectory=traj, source=source_test)
    print("Elapsed time in RadiationFactory: ",time.time()-t0)

    print('Screen distance :')
    print(Rad.distance)

    print("screen shape ")
    print(Rad.intensity.shape)

    print('X max :')
    print(Rad.X.max())

    print('Y max :')
    print(Rad.Y.max())

    print('intensity max ()')
    print(Rad.max())

    print('plot')
    Rad.plot(title="FAR FIELD")


def Exemple_NEARFIELD():
    from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane as Undulator
    from pySRU.ElectronBeam import ElectronBeam
    from pySRU.SourceUndulatorPlane import SourceUndulatorPlane
    import time


    undulator_test = Undulator(K=1.87, period_length=0.035, length=0.035 * 14)
    electron_beam_test = ElectronBeam(Electron_energy=1.3, I_current=1.0)
    magnetic_field_test = undulator_test.create_magnetic_field()
    source_test = SourceUndulatorPlane(undulator=undulator_test,
                         electron_beam=electron_beam_test,
                         magnetic_field=magnetic_field_test)

    traj = TrajectoryFactory(Nb_pts=2000, method=TRAJECTORY_METHOD_ODE).create_from_source(source_test)

    t0 = time.time()
    Rad = RadiationFactory(photon_frequency=source_test.harmonic_frequency(1), method=RADIATION_METHOD_NEAR_FIELD, Nb_pts=101
                           ).create_for_one_relativistic_electron(trajectory=traj, source=source_test,distance=100)
    print("Elapsed time in RadiationFactory: ",time.time()-t0)

    print('Screen distance :')
    print(Rad.distance)

    print("screen shape ")
    print(Rad.intensity.shape)

    print('X max :')
    print(Rad.X.max())

    print('Y max :')
    print(Rad.Y.max())

    print('intensity max ()')
    print(Rad.max())

    print('plot')
    Rad.plot()

def Exemple_APPROX():
    from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane as Undulator
    from pySRU.ElectronBeam import ElectronBeam
    from pySRU.SourceUndulatorPlane import SourceUndulatorPlane
    import time


    undulator_test = Undulator(K=1.87, period_length=0.035, length=0.035 * 14)
    electron_beam_test = ElectronBeam(Electron_energy=1.3, I_current=1.0)
    magnetic_field_test = undulator_test.create_magnetic_field()
    source_test = SourceUndulatorPlane(undulator=undulator_test,
                         electron_beam=electron_beam_test,
                         magnetic_field=magnetic_field_test)

    traj = TrajectoryFactory(Nb_pts=2000, method=TRAJECTORY_METHOD_ODE).create_from_source(source_test)

    t0 = time.time()
    Rad = RadiationFactory(photon_frequency=source_test.harmonic_frequency(1), method=RADIATION_METHOD_APPROX, Nb_pts=101
                           ).create_for_one_relativistic_electron(trajectory=traj, source=source_test,distance=100)
    print("Elapsed time in RadiationFactory: ",time.time()-t0)


    print('Screen distance :')
    print(Rad.distance)

    print("screen shape ")
    print(Rad.intensity.shape)

    print('X max :')
    print(Rad.X.max())

    print('Y max :')
    print(Rad.Y.max())

    print('intensity max ()')
    print(Rad.max())

    print('plot')
    Rad.plot()

if __name__ == "__main__" :

    Exemple_FARFIELD()
    Exemple_APPROX()
    Exemple_NEARFIELD()







# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
from  scipy.constants import physical_constants
#from mpl_toolkits.mplot3d import Axes3D
import scipy.constants as codata
import time
from scipy.interpolate import interp1d
from pySRU.Radiation import Radiation
from pySRU.MagneticField import MagneticField
from pySRU.SourceUndulatorPlane import SourceUndulatorPlane
from pySRU.SourceBendingmagnet import SourceBendingMagnet
from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane as Undulator
from pySRU.MagneticStructureBendingMagnet import MagneticStructureBendingMagnet as BM

from pySRU.TrajectoryFactory import TrajectoryFactory, TRAJECTORY_METHOD_ANALYTIC,TRAJECTORY_METHOD_ODE

from pySRU.RadiationFactory import RadiationFactory,RADIATION_METHOD_NEAR_FIELD, \
                                 RADIATION_METHOD_APPROX_FARFIELD


eV_to_J = codata.e

class Simulation(object):

    def __init__(self, source, trajectory_fact, radiation_fact, trajectory, radiation, electric_field=None):
        self.source=source
        self.trajectory_fact=trajectory_fact
        self.radiation_fact=radiation_fact
        self.trajectory=trajectory
        self.radiation=radiation
        self.electric_field=electric_field

# faire un autre constructeur pour la copie


    def copy(self):
        return Simulation(source=self.source.copy(), trajectory_fact=self.trajectory_fact.copy(),
                                   radiation_fact=self.radiation_fact.copy(),
                                   trajectory=self.trajectory.copy(), radiation=self.radiation.copy())


    def update_radiation(self):
        self.radiation.intensity = self.radiation_fact.calculate_radiation_intensity(trajectory=self.trajectory,
                                                                                     source=self.source,
                                                                                     distance=self.radiation.distance,
                                                                                     X_array=self.radiation.X,
                                                                                     Y_array=self.radiation.Y)
    #change

    def change_distance(self,D,update_radiation=1):
        self.radiation.distance=D
        if update_radiation:
            self.update_radiation()

    def change_radiation_method(self,method,update_radiation=1):
        self.radiation_fact.method=method
        if update_radiation:
            self.update_radiation()

    def change_trajectory_method(self,method,update_radiation=1) :
        self.trajectory_fact.method = method
        self.trajectory =self.trajectory_fact.create_from_source(source=self.source)
        if update_radiation:
            self.update_radiation()


    def change_photon_frequency(self, omega, update_radiation=1) :
        self.radiation_fact.photon_frequency=omega
        if update_radiation:
            self.update_radiation()

    def change_harmonic_number(self,harmonic_number,update_radiation=True) :
        self.radiation_fact.photon_frequency= harmonic_number * self.source.harmonic_frequency(1)
        if update_radiation:
            self.update_radiation()

    def change_energy_eV(self,E,update_radiation=1) :
        omega = E * eV_to_J / codata.hbar
        self.change_photon_frequency(omega)
        if update_radiation:
            self.update_radiation()



    def change_Nb_pts_trajectory(self, Nb_pts,update_radiation=1):
        self.trajectory_fact.Nb_pts = int(Nb_pts)

        self.trajectory = self.trajectory_fact.create_from_source(source=self.source)
        if update_radiation:
            self.update_radiation()

    def change_nb_period(self,Nb_period,update_radiation=1):
        self.source.magnetic_structure.length=Nb_period*self.source.magnetic_structure.period_length
        self.source.magnetic_field=self.source.magnetic_structure.create_magnetic_field()
        self.trajectory_fact.Nb_pts = int(self.source.choose_nb_pts_trajectory(2))
        self.trajectory = self.trajectory_fact.create_from_source(source=self.source)
        if update_radiation:
            self.update_radiation()

    def change_Nb_pts_radiation(self,Nb_pts,update_radiation=1):
        self.radiation_fact.Nb_pts = Nb_pts
        self.radiation.change_Nb_pts(Nb_pts)
        if update_radiation:
            self.update_radiation()

    def change_XY_radiation(self,X,Y,update_radiation=1):
        self.radiation.X=X
        self.radiation.Y=Y
        if update_radiation:
            self.update_radiation()

    #TODO possible que pour Undulator, a revoir , a tester
    def calculate_on_ring_number(self, ring_number_max,update_radiation=1,Nb_pts=101):
        harmonic_number=np.floor(self.radiation_fact.photon_frequency / self.source.harmonic_frequency(1))
        self.radiation_fact.photon_frequency=self.source.harmonic_frequency(harmonic_number)
        X=np.array([])
        Y = np.array([])
        t = np.linspace(0.0, 2.0* np.pi, Nb_pts)
        for i in range(int(ring_number_max)+1) :
            Xi,Yi= self.source.describe_ring(distance=self.radiation.distance,harmonic_number=harmonic_number,
                                              ring_number=i,t=t)
            X=np.concatenate((X,Xi))
            Y=np.concatenate((Y,Yi))
        self.change_XY_radiation(X=X,Y=Y,update_radiation=update_radiation)


    def calculate_on_central_cone(self,harmonic_number=None,npoints_x=101,npoints_y=101,update_radiation=1):
        if harmonic_number == None :
            harmonic_number=np.floor(self.radiation_fact.photon_frequency/self.source.harmonic_frequency(1))
        else :
            self.change_harmonic_number(harmonic_number, update_radiation=0)
        theta_max=self.source.angle_deflection_central_cone(harmonic_number)
        if self.radiation.distance==None :
            X_max=theta_max
            Y_max=theta_max

        else :
            X_max=self.radiation.distance*theta_max
            Y_max=self.radiation.distance*theta_max


        # if npoints_x is None:
        #     npoints_x = self.radiation_fact.Nb_pts
        # if npoints_y is None:
        #     npoints_y = self.radiation_fact.Nb_pts

        X=np.linspace(0.0,X_max,npoints_x)
        Y=np.linspace(0.0,Y_max,npoints_y)
        # Y,X=np.meshgrid(X,Y)

        X_ones = np.ones_like(X)
        Y_ones = np.ones_like(Y)
        X = np.outer(X,Y_ones)
        Y = np.outer(X_ones,Y)

        self.change_XY_radiation(X=X,Y=Y,update_radiation=update_radiation)


    def calculate_until_ring_number(self,ring_number,harmonic_number=1, XY_are_list=True,update_radiation=1,
                                    npoints=None):
        # harm_num=self.radiation_fact.photon_frequency/self.source.harmonic_frequency(1)
        if npoints != None:
            self.radiation_fact.Nb_pts = npoints

        self.change_harmonic_number(harmonic_number,update_radiation=0)
        observation_angle = np.linspace(0.0, self.source.angle_ring_number(harmonic_number=harmonic_number,
                                        ring_number=ring_number),self.radiation_fact.Nb_pts)
        self.calculate_for_observation_angles(observation_angle,XY_are_list=XY_are_list,update_radiation=update_radiation)



    def calculate_for_observation_angles(self,observation_angle,XY_are_list=True,Nb_pts=101,update_radiation=1):
        D=self.radiation.distance
        if D==None :
            D=1
        if XY_are_list :
            X = np.array([])
            Y = np.array([])
            t = np.linspace(0.0, 2.0 * np.pi, Nb_pts)
            for theta in observation_angle:
                if theta==0.0 :
                    Xi=np.array([0.0])
                    Yi=np.array([0.0])
                else :
                    Xi=np.cos(t)*D*theta
                    Yi=np.sin(t)*D*theta
                X = np.concatenate((X, Xi))
                Y = np.concatenate((Y, Yi))
        else :
            X=D*observation_angle
            Y=D*observation_angle
            # X,Y=np.meshgrid(X,Y)
            X_ones = np.ones_like(X)
            Y_ones = np.ones_like(Y)
            X = np.outer(X,Y_ones)
            Y = np.outer(X_ones,Y)

        self.change_XY_radiation(X=X,Y=Y,update_radiation=update_radiation)


    #######
    # spectrum
    #######

    def spectrum(self, omega_array=None, is_quadrant=0):
        if omega_array is None:
            omega1 = self.source. choose_photon_frequency()
            omega_array1 = np.arange(omega1 * 0.8, omega1 * 1.11, omega1 * 0.02)
            omega_array3 = np.arange(omega1 * 1.8, omega1 * 2.11, omega1 * 0.02)
            omega_array5 = np.arange(omega1 * 2.8, omega1 * 3.11, omega1 * 0.02)

            omega_array2=np.arange(omega1*1.12,omega1*1.79,omega1*0.4)
            omega_array4=np.arange(omega1*2.12,omega1*2.79,omega1*0.4)

            omega_array = np.concatenate((omega_array1,omega_array2,omega_array3,omega_array4,omega_array5))

        print("Calculating spectrum for %d energy points..."%(omega_array.size))
        spectrum = np.zeros_like(omega_array)

        for i in range(len(spectrum)):
            print("Point %d of %d..."%(1+i,spectrum.size))
            self.change_photon_frequency(omega_array[i],update_radiation=1)
            spectrum[i] = self.radiation.integration(is_quadrant=is_quadrant)

        return spectrum, omega_array

    def spectrum_on_radiation_max(self, omega_array=None):
        if omega_array == None:
            omega1 = self.source.omega1()
            omega_array = np.arange(omega1 * 0.9, 5.0 * omega1 * 1.1, omega1 * 0.01)
        spectrum = np.zeros_like(omega_array)
        print(len(spectrum))
        for i in range(len(spectrum)):
            print(i)
            self.change_photon_frequency(omega_array[i])
            spectrum[i] = self.radiation.max()
        return spectrum, omega_array

    def spectrum_on_axis(self, omega_array=None):
        save_X=self.radiation.X
        X=np.array([0.0])
        save_Y=self.radiation.Y
        Y=np.array([0.0])
        self.change_XY_radiation(X=X,Y=Y)
        if omega_array == None:
            omega1 = self.source.harmonic_frequency(1)
            omega_array = np.arange(omega1 * 0.9, 5.0 * omega1 * 1.1, omega1 * 0.001)
        spectrum = np.zeros_like(omega_array)
        for i in range(len(spectrum)):
            self.change_photon_frequency(omega_array[i])
            spectrum[i] = self.radiation.intensity.max()
        self.change_XY_radiation(X=save_X,Y=save_Y)
        return spectrum, omega_array


    def calculate_spectrum_on_slit(self,abscissas_array=None,use_eV=0,is_quadrant=0,do_plot=1):

        if use_eV:
            conversion_factor = codata.hbar / eV_to_J
            xlabel = "Photon energy [eV]"
        else:
            conversion_factor = 1.0
            xlabel = "Photon frequency [Hz]"

        if abscissas_array is None:
            omega_array = None
        else:
            omega_array = abscissas_array / conversion_factor



        spectrum,omega_array=self.spectrum(omega_array=omega_array,is_quadrant=is_quadrant)

        if do_plot:
            import matplotlib.pyplot as plt
            plt.plot(conversion_factor*omega_array, spectrum, label='calculated radiation on slit')
            plt.legend()
            plt.xlabel(xlabel)
            plt.ylabel("Flux [phot/s/0.1%bw]")
            plt.show()

        print("Maximum flux is %g"%spectrum.max())
        return conversion_factor*omega_array,spectrum



    #TODO possible que pour undultor et c'est le quart du cone central !
    def calculate_spectrum_central_cone(self,harmonic_number=1,theoretical_value=False,abscissas_array=None,use_eV=0,do_plot=1):


        if use_eV:
            conversion_factor = codata.hbar / eV_to_J
            xlabel = "Photon energy [eV]"
        else:
            conversion_factor = 1.0
            xlabel = "Photon frequency [Hz]"

        if abscissas_array is None:
            omega_array = None
        else:
            omega_array = abscissas_array / conversion_factor

        #self.radiation_fact.distance=None
        theta_max = self.source.angle_deflection_central_cone()
        if self.radiation.distance == None:
            X_max = theta_max
            Y_max = theta_max
        else:
            X_max = self.radiation.distance * theta_max
            Y_max = self.radiation.distance * theta_max

        X = np.linspace(0.0, X_max, 21)
        Y = np.linspace(0.0, Y_max, 21)
        # XY = np.meshgrid(X,Y)
        XX = np.outer(X,np.ones_like(Y))
        YY = np.outer(np.ones_like(X),Y)
        self.change_XY_radiation(XX,YY,update_radiation=0)
        # if do_plot:
        #     self.radiation.plot()


        # main calculation in one quadrant
        if omega_array is None: # pre-calculate default
            omega1 = self.source.harmonic_frequency(1)
            omega_array = np.linspace(omega1*harmonic_number*0.95,omega1*harmonic_number*1.05,30)

        spectrum,omega_array=self.spectrum(omega_array=omega_array,is_quadrant=1)

        if do_plot:
            import matplotlib.pyplot as plt

            if theoretical_value :
                spectrum_theoretical=self.source.theoretical_flux_integrated_central_cone(harmonic_number)
                plt.plot(conversion_factor*harmonic_number*omega1, spectrum_theoretical, "g*", label='theoretical value')
                print("Theoretical flux",spectrum_theoretical,"Max spectrum: ",spectrum.max())
            plt.plot(conversion_factor*omega_array, spectrum, label='calculad radiation on central cone for harmonic %d'%harmonic_number)
            plt.legend()
            plt.xlabel(xlabel)
            plt.ylabel("Flux [phot/s/0.1%bw]")
            plt.show()

        return conversion_factor*omega_array,spectrum




    def calculate_spectrum_on_axis(self, omega_array=None, title="On axis flux", use_eV=0, do_plot=1):

        if use_eV:
            conversion_factor = codata.hbar / eV_to_J
            xlabel = "Photon energy [eV]"
        else:
            conversion_factor = 1.0
            xlabel = "Photon frequency [Hz]"

        spectrum,omega_array=self.spectrum_on_axis(omega_array=omega_array)

        if do_plot:

            # compute theoretical spectrum (to compare with)
            omega1 = self.source.harmonic_frequency(1)
            harm_num_min = np.ceil(omega_array.min() / omega1)
            harm_num_max = np.floor(omega_array.max() / omega1)
            harm_number=np.arange(harm_num_min,harm_num_max+0.5,1)

            spectrum_theoretical_scan = self.source.theoretical_flux_on_axis_omega_scan(harm_number,omega_array)
            # not needed
            spectrum_theoretical=np.zeros_like(harm_number)
            for i in range(len(spectrum_theoretical)):
                spectrum_theoretical[i]=self.source.theoretical_flux_on_axis(harm_number[i])

            # plot
            import matplotlib.pyplot as plt
            plt.plot(conversion_factor*omega_array,spectrum,label='calculated radiation on axis')
            plt.plot(conversion_factor*omega_array,spectrum_theoretical_scan,label='theoretical radiation on axis')
            plt.plot(conversion_factor*harm_number*omega1,spectrum_theoretical, "g*",label='theoretical peak value')
            plt.legend()
            plt.xlabel(xlabel)
            plt.ylabel("Flux (phot/s/0.1%bw/mrad2")
            plt.title(title)
            plt.show()

        return conversion_factor*omega_array,spectrum
    # plot

    def plot_magnetic_field_along_Z(self,X=None,Y=None):
        Z=self.trajectory.t*self.source.electron_speed()*codata.c
        if Y==None :
            Y=self.trajectory_fact.initial_condition[4]
        if X == None:
            X = self.trajectory_fact.initial_condition[3]
        self.source.magnetic_field.plot_z(X=X,Y=Y,Z=Z)

    def print_parameters(self):
        self.source.print_parameters()

        self.trajectory_fact.print_parameters()

        self.radiation_fact.print_parameters()

        print(" Others")
        print("     Distance: ",self.radiation.distance)
        if (type(self.source)==BM) :
            print("     Central Cone width (urad) for  n=1: %6.3f, n=3: %6.3f, n=5: %6.3f "%
                  ( 1e6 * self.source.angle_deflection_central_cone(harmonic_number=1),
                    1e6 * self.source.angle_deflection_central_cone(harmonic_number=3),
                    1e6 * self.source.angle_deflection_central_cone(harmonic_number=5) ))
            print("     First ring position (urad) for n=1: %6.3f, n=3: %6.3f, n=5: %6.3f "%
                  ( 1e6 * self.source.angle_ring_number(harmonic_number=1,ring_number=1),
                    1e6 * self.source.angle_ring_number(harmonic_number=3,ring_number=1),
                    1e6 * self.source.angle_ring_number(harmonic_number=5,ring_number=1) ))

    def plot_everything(self):
        self.trajectory.plot()
        self.trajectory.plot_3D()
        self.radiation.plot()
        if (type(self.source) == BM):
            self.calculate_on_central_cone()
        self.radiation.plot()
        self.plot_spectrum()
        self.calculate_spectrum_on_axis()


    #TODO a tester attention possible que pour BENDING magnet
    def create_theoretical_radiation(self):
        #TODO faire le cas None
        radiation=self.radiation.copy()
        shape1 = radiation.intensity.shape
        # X = radiation.X.flatten()
        # Y = radiation.X.flatten()
        # res=radiation.intensity.flatten()
        # cas grid :
        print(type(radiation))
        print(type(radiation.Y))
        Y=radiation.Y
        X=radiation.X
        for i in range(Y.shape[0]):
            for j in range(Y.shape[1]) :
                observation_angle=(Y[i,j]/radiation.distance)
                #TODO changer pour qu'on calcul avec des angles
                radiation.intensity[i,j]=self.source.radiation_theoretical(
                    omega=self.radiation_fact.photon_frequency,observation_angle=observation_angle)
        return radiation


    def create_theoretical_radiation2(self):
        radiation = self.radiation.copy()
        shape1 = radiation.intensity.shape
        # X = radiation.X.flatten()
        # Y = radiation.X.flatten()
        # res=radiation.intensity.flatten()
        # cas grid :
        print(type(radiation))
        print(type(radiation.Y))
        Y = radiation.Y
        X = radiation.X
        for i in range(Y.shape[0]):
            for j in range(Y.shape[1]):
                observation_angle = (Y[i, j] / radiation.distance)
                radiation.intensity[i, j] = self.source.radiation_theoretical(
                    omega=self.radiation_fact.photon_frequency, observation_angle=observation_angle)
        return radiation

#


    def error_radiation_method_nb_pts_traj(self, method, nb_pts):
        sim2 = self.copy()
        sim2.change_radiation_method(method)
        error = np.zeros_like(nb_pts)
        print(len(nb_pts))
        for i in range(len(nb_pts)):
            print(i)
            self.change_Nb_pts_trajectory(nb_pts[i])
            sim2.change_Nb_pts_trajectory(nb_pts[i])
            error[i] = self.radiation.error_max(sim2.radiation)
        return error

    def error_radiation_nb_pts_traj(self, good_value, nb_pts):
        error = np.zeros_like(nb_pts)
        print(len(nb_pts))
        for i in range(len(nb_pts)):
            print(i)
            self.change_Nb_pts_trajectory(nb_pts[i])
            error[i] = (np.abs(good_value-self.radiation.intensity)).max()
        return error


    def error_radiation_method_distance(self, method, D):
        sim2 = self.copy()
        sim2.radiation_fact.method = method
        error = np.zeros_like(D)
        print(len(D))
        for i in range(len(D)):
            print(i)
            self.change_distance(D[i])
            sim2.change_distance(D[i])
            error[i] = self.radiation.error_max(sim2.radiation)
        return error


    def error_trajectory_method(self, method, nb_pts):
        sim2 = self.copy()
        sim2.change_trajectory_method(method)
        error_rad = np.zeros_like(nb_pts)
        error_traj = np.zeros((10, len(nb_pts)))
        for i in range(len(nb_pts)):
            print(i)
            self.change_Nb_pts_trajectory(nb_pts[i])
            sim2.change_Nb_pts_trajectory(nb_pts[i])
            error_traj[:, i] = self.trajectory.error_max(sim2.trajectory)
            error_rad[i] = self.radiation.error_max(sim2.radiation)
        traj_error_traj = self.trajectory_fact.create_from_array(error_traj)
        return error_rad, traj_error_traj


    def relativ_error_radiation_method_distance(self, method, D):
        sim2 = self.copy()
        sim2.change_radiation_method(method)
        error_relativ = np.array(len(D))
        for i in range(len(D)):
            self.change_distance(D[i])
            sim2.change_distance(D[i])
            error_relativ[i] = self.radiation.relativ(sim2.radiation)
        return error_relativ


def create_simulation(magnetic_structure,electron_beam, magnetic_field=None, photon_energy=None,
                      traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                      rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                      initial_condition=None, distance=None,XY_are_list=False,X=None,Y=None) :

    if type(magnetic_structure)==Undulator :
        source=SourceUndulatorPlane(undulator=magnetic_structure,
                                    electron_beam=electron_beam, magnetic_field=magnetic_field)
        print("Calculating undulator source...")
    elif type(magnetic_structure)==BM:
        source = SourceBendingMagnet(magnetic_structure=magnetic_structure,
                                      electron_beam=electron_beam, magnetic_field=magnetic_field)
        print("Calculating bending magnet source...")
    else :
        raise Exception('magnet type unknown')

    if photon_energy==None :
        omega=source.choose_photon_frequency()
    else :
        omega = photon_energy * eV_to_J / codata.hbar

    #
    # XY grid
    #
    if distance==None and (rad_method==RADIATION_METHOD_NEAR_FIELD) :
        distance=source.choose_distance_automatic(2)

    if Nb_pts_trajectory==None :
        Nb_pts_trajectory = int(source.choose_nb_pts_trajectory(0.01,photon_frequency=omega))

    if X is None or Y is None :
        if (X != None) :
            Y=X
        elif Y != None :
            X=Y
        else :
            theta_max=source.choose_angle_deflection_max()
            if distance==None :
                X_max=theta_max
                Y_max=theta_max
            else :
                X_max = distance * theta_max
                Y_max = distance * theta_max
            X = np.linspace(0.0, X_max, Nb_pts_radiation)
            Y = np.linspace(0.0, Y_max, Nb_pts_radiation)

    if type(X) == float:
        X= np.linspace(0.0, X, Nb_pts_radiation)
    if type(Y) == float:
        Y = np.linspace(0.0, Y, Nb_pts_radiation)

    # if X.shape != Y.shape :
    #     raise Exception('X and Y must have the same shape')
    Nb_pts_radiation = X.size # len(X.flatten())

    #print('step 1')
    traj_fact=TrajectoryFactory(Nb_pts=Nb_pts_trajectory,method=traj_method,initial_condition=initial_condition)

    if (traj_fact.initial_condition == None):
        # print('crearte initial cond automat')
        traj_fact.initial_condition = source.choose_initial_contidion_automatic()



    #print('step 2')

    rad_fact=RadiationFactory(method=rad_method, photon_frequency=omega)


    #print('step 3')
    trajectory=traj_fact.create_from_source(source=source)
    #print('step 4')
    # print("input: ", XY_are_list)
    # print("input: ", X.shape)
    # print("input: ", Y.shape)
    if False: # before storing electric field
        radiation = rad_fact.create_for_one_relativistic_electron(trajectory=trajectory, source=source, XY_are_list=XY_are_list,
                                                                  distance=distance, X=X, Y=Y)

        print("radiation: ", radiation.intensity.shape)
        print("radiation: ", radiation.X.shape)
        print("radiation: ", radiation.Y.shape)
        print("radiation: ", radiation.distance)
    else: # when also storing electric field
        if XY_are_list:
            electric_field = rad_fact.calculate_electrical_field(
                trajectory, source, X, Y, distance)
            EEx = electric_field._electrical_field[:, 0]
            EEy = electric_field._electrical_field[:, 1]
            EEz = electric_field._electrical_field[:, 2]
            radiation2 = Radiation(np.abs(EEx) ** 2 + np.abs(EEy) ** 2 + np.abs(EEz) ** 2, X, Y, distance)
        else:
            XX = np.outer(X, np.ones_like(Y))
            YY = np.outer(np.ones_like(X), Y)
            electric_field = rad_fact.calculate_electrical_field(
                trajectory, source,
                XX,
                YY,
                distance)
            EEx = electric_field._electrical_field[:, :, 0]
            EEy = electric_field._electrical_field[:, :, 1]
            EEz = electric_field._electrical_field[:, :, 2]
            radiation2 = Radiation(np.abs(EEx)** 2 + np.abs(EEy)**2 + np.abs(EEz)**2, XX, YY, distance)

        # print("efield: ", electric_field._electrical_field.shape)
        # print("efield: ", electric_field._X.shape)
        # print("efield: ", electric_field._Y.shape)
        # print("efield: ", electric_field._distance)

    #print('step 5')
    return Simulation(source=source,
                      trajectory_fact=traj_fact,
                      radiation_fact=rad_fact,
                      trajectory=trajectory,
                      radiation=radiation2,
                      electric_field=electric_field)

# exemples

def Example_minimum():

    from  pySRU.ElectronBeam import ElectronBeam

    print("======================================================================")
    print("======      Undulator from X-ray data booklet                  =======")
    print("====== fig 2.5 in  http://xdb.lbl.gov/Section2/Sec_2-1.html    =======")
    print("======================================================================")

    # note that the flux in the reference fig 2.6 is a factor 10 smaller than the calculated here.
    # This factor comes from the units:
    #     here: phot / s  / A / 0.1%bw / (mrad)^2
    #     ref : phot / s  / A /   1%bw / (0.1 mrad)^2

    undulator_test = Undulator(K=1.87, period_length=0.035, length=0.035 * 14)
    electron_beam_test = ElectronBeam(Electron_energy=1.3, I_current=1.0)

    simulation_test = create_simulation(magnetic_structure=undulator_test,electron_beam=electron_beam_test,
                        magnetic_field=None, photon_energy=None,
                        traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                        rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                        initial_condition=None, distance=None,XY_are_list=False,X=None,Y=None)


    simulation_test.print_parameters()

    simulation_test.trajectory.plot_3D(title="Electron Trajectory")

    simulation_test.radiation.plot(title="Flux in far field vs angle")

    #  second calculation, change distance and also change the grid to accept the central cone
    simulation_test.change_distance(D=100)
    simulation_test.calculate_on_central_cone()

    simulation_test.radiation.plot(title="New distance: %3.1f m"%simulation_test.radiation.distance)



def Example_meshgrid():
    from pySRU.ElectronBeam import ElectronBeam

    print("======================================================================")
    print("======      Undulator U18 from ESRF with K=1.68                =======")
    print("======================================================================")

    beam_ESRF = ElectronBeam(Electron_energy=6.0, I_current=0.2)
    ESRF18 = Undulator(K=1.68, period_length=0.018, length=2.0)

    #
    # radiation in a defined meah
    #
    print('create radiation a given screen (40mm x 40mm @ 100 m )')
    X = np.linspace(-0.02,0.02,101)
    Y = np.linspace(-0.02,0.02,101)
    distance = 100.0
    simulation_test = create_simulation(magnetic_structure=ESRF18,electron_beam=beam_ESRF,
                        magnetic_field=None, photon_energy=None,
                        traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                        rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                        initial_condition=None, distance=distance,XY_are_list=False,X=X,Y=Y)

    simulation_test.print_parameters()
    simulation_test.trajectory.plot_2D()
    simulation_test.radiation.plot(title=" radiation in a defined screen (40mm x 40mm @ 100 m )")

    #
    # up to a maximum X and Y
    #
    print('create simulation for a given maximum X and Y ')
    simulation_test = create_simulation(magnetic_structure=ESRF18, electron_beam=beam_ESRF,
                                        traj_method=TRAJECTORY_METHOD_ANALYTIC,
                                        rad_method=RADIATION_METHOD_APPROX_FARFIELD,
                                        distance=100,X=0.01,Y=0.01)

    simulation_test.radiation.plot(title='simulation for a maximum X=0.01 and Y=0.01')


def Example_meshgrid_on_central_cone_and_rings():
    from pySRU.ElectronBeam import ElectronBeam

    print("======================================================================")
    print("======      Undulator U18 from ESRF with K=1.68                =======")
    print("======================================================================")

    beam_ESRF = ElectronBeam(Electron_energy=6.0, I_current=0.2)
    ESRF18 = Undulator(K=1.68, period_length=0.018, length=2.0)

    #
    # radiation in a defined mesh with only one point to save time
    #

    distance = None
    simulation_test = create_simulation(magnetic_structure=ESRF18,electron_beam=beam_ESRF,
                        magnetic_field=None, photon_energy=None,
                        traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                        rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                        initial_condition=None, distance=distance,XY_are_list=False,X=np.array([0]),Y=np.array([0]))


    #
    # central cone
    #
    harmonic_number = 1
    print('create radiation in a screen including central cone for harmonic %d'%harmonic_number)
    simulation_test.calculate_on_central_cone(harmonic_number=harmonic_number,npoints_x=50,npoints_y=50)
    simulation_test.print_parameters()
    simulation_test.radiation.plot(title=("radiation in a screen including central cone, harmonic %d,at D="+repr(distance))%harmonic_number)

    #
    # up to a given ring
    #
    harmonic_number = 1
    ring_number = 1
    print('create radiation a given screen including ring %d for harmonic %d'%(ring_number,harmonic_number))
    simulation_test.calculate_until_ring_number(harmonic_number=harmonic_number, ring_number=ring_number, XY_are_list=False, npoints=51)
    simulation_test.print_parameters()
    simulation_test.radiation.plot(title=" radiation in a screen containing harmonic %d up to ring %d ring"%
                                         (harmonic_number,ring_number))





def Example_spectrum_on_axis():
    from pySRU.ElectronBeam import ElectronBeam

    print("======================================================================")
    print("======      Undulator U18 from ESRF with K=1.68                =======")
    print("======================================================================")

    beam_ESRF = ElectronBeam(Electron_energy=6.0, I_current=0.2)
    ESRF18 = Undulator(K=1.68, period_length=0.018, length=2.0)

    # note that distance=None is important to get results in angles and therefore flux in ph/s/0.1%bw/mrad2
    simulation_test = create_simulation(magnetic_structure=ESRF18,electron_beam=beam_ESRF,
                        magnetic_field=None, photon_energy=None,
                        traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                        rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                        initial_condition=None, distance=None,XY_are_list=False,X=np.array([0]),Y=np.array([0]))


    x,y = simulation_test.calculate_spectrum_on_axis(use_eV=1,do_plot=1)

    # dump file
    filename = "spectrum.spec"
    f = open(filename,'w')
    f.write("#F %s\n\n#S 1 undulator flux using pySRU\n"%filename)
    f.write("#N 2\n#L Photon energy [eV]  Flux on axis [phot/s/0.1%bw/mrad2]\n")
    for i in range(x.size):
        f.write("%f  %g\n"%(x[i],y[i]))
    f.close()
    print("File written to disk: %s"%filename)

def Example_spectrum_on_slit():
    from pySRU.ElectronBeam import ElectronBeam

    print("======================================================================")
    print("======      Undulator U18 from ESRF with K=1.68                =======")
    print("======================================================================")

    beam_ESRF = ElectronBeam(Electron_energy=6.0, I_current=0.2)
    ESRF18 = Undulator(K=1.68, period_length=0.018, length=2.0)

    print('create radiation a given screen (.025 x .025 mrad) accepting central cone')
    is_quadrant = 1
    distance = None # using angles

    if is_quadrant:
        X = np.linspace(0,0.0125e-3,26)
        Y = np.linspace(0,0.0125e-3,26)
    else:
        X = np.linspace(-0.0125e-3,0.0125e-3,51)
        Y = np.linspace(-0.0125e-3,0.0125e-3,51)

    if distance != None:
        X *= distance
        Y *= distance

    simulation_test = create_simulation(magnetic_structure=ESRF18,electron_beam=beam_ESRF,
                        magnetic_field=None, photon_energy=None,
                        traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                        rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                        initial_condition=None, distance=distance,XY_are_list=False,X=X,Y=Y)



    simulation_test.print_parameters()
    simulation_test.radiation.plot(title=("radiation in a screen for first harmonic"))
    print("Integrated flux at resonance: %g photons/s/0.1bw"%(simulation_test.radiation.integration(is_quadrant=is_quadrant)))

    x,y = simulation_test.calculate_spectrum_on_slit(abscissas_array=None,use_eV=1,is_quadrant=is_quadrant)

    # dump file
    filename = "spectrum.spec"
    f = open(filename,'w')
    f.write("#F %s\n\n#S 1 undulator flux using pySRU\n"%filename)
    f.write("#N 2\n#L Photon energy [eV]  Flux on slit [phot/s/0.1%bw]\n")
    for i in range(x.size):
        f.write("%f  %g\n"%(x[i],y[i]))
    f.close()
    print("File written to disk: %s"%filename)


def Example_spectrum_on_central_cone():
    from pySRU.ElectronBeam import ElectronBeam

    print("======================================================================")
    print("======      Undulator U18 from ESRF with K=1.68                =======")
    print("======================================================================")

    beam_ESRF = ElectronBeam(Electron_energy=6.0, I_current=0.2)
    ESRF18 = Undulator(K=1.68, period_length=0.018, length=2.0)


    simulation_test = create_simulation(magnetic_structure=ESRF18,electron_beam=beam_ESRF,
                        magnetic_field=None, photon_energy=None,
                        traj_method=TRAJECTORY_METHOD_ANALYTIC,Nb_pts_trajectory=None,
                        rad_method=RADIATION_METHOD_APPROX_FARFIELD, Nb_pts_radiation=101,
                        initial_condition=None, distance=None,XY_are_list=False,X=np.array([0]),Y=np.array([0]))

    harmonic_number = 1
    x,y = simulation_test.calculate_spectrum_central_cone(harmonic_number=harmonic_number,theoretical_value=True,abscissas_array=None,use_eV=1)

    # dump file
    filename = "spectrum.spec"
    f = open(filename,'w')
    f.write("#F %s\n\n#S 1 undulator flux using pySRU\n"%filename)
    f.write("#N 2\n#L Photon energy [eV]  Flux on central cone for harmonic %d [phot/s/0.1bw]\n"%harmonic_number)
    for i,xi in enumerate(x):
        f.write("%f  %g\n"%(xi,y[i]))
    f.close()
    print("File written to disk: %s"%filename)

# TODO ne marche plus  qd on rentre a la main l'energy du photon?

def Example_list():
    from pySRU.ElectronBeam import ElectronBeam

    beam_ESRF = ElectronBeam(Electron_energy=6.0, I_current=0.2)
    ESRF18 = Undulator(K=1.68, period_length=0.018, length=2.0)

    X = np.linspace(0.0, 0.0002,1001)
    Y = np.linspace(0.0, 0.0002, 1001)
    simulation_test = create_simulation(magnetic_structure=ESRF18,electron_beam=beam_ESRF,
                                        X=X,Y=Y,XY_are_list=True)

    simulation_test.print_parameters()

    simulation_test.trajectory.plot_3D()

    simulation_test.radiation.plot()

    simulation_test.calculate_on_ring_number(ring_number_max=2)

    simulation_test.radiation.plot()

    simulation_test.radiation.plot_ring()


    observation_angle = np.linspace(0.0, simulation_test.source.angle_ring_number(1, 2), 51)

# TODO verifier quelle marche bien
    simulation_test.calculate_for_observation_angles(XY_are_list=True,observation_angle=observation_angle)

    simulation_test.radiation.plot()



if __name__ == "__main__" :

    pass
    # Example_minimum()
    # Example_meshgrid()
    # Example_meshgrid_on_central_cone_and_rings()
    # Example_spectrum_on_axis()
    # Example_spectrum_on_slit()
    # Example_spectrum_on_central_cone()
    # Example_list()





# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
import scipy.constants as codata
from scipy.special import jn,yn,jv,yv
from pySRU.MagneticField import MagneticField
from pySRU.MagneticStructure import MagneticStructure , PLANE_UNDULATOR,BENDING_MAGNET



class MagneticStructureUndulatorPlane(MagneticStructure):
    def __init__(self, K, period_length, length):
        super(self.__class__, self).__init__(magnet_type=PLANE_UNDULATOR)
        self.K = K
        self.period_length = period_length
        self.length=length

    def copy(self):
        return MagneticStructureUndulatorPlane(K=self.K, period_length=self.period_length, length=self.length)

    def fct_magnetic_field(self, z, y, x, harmonic_number, coordonnee='y'):
        lambda_h = self.period_length / harmonic_number
        ku = 2.0 * np.pi / self.period_length

        if coordonnee == 'x':
            dB = 0.0
        else:
            # codata.m_e * codata.c / codata.e= 0.00170450894933
            K_to_B = ku * codata.m_e * codata.c / codata.e
            if coordonnee == 'y':

                Bo = (self.K * K_to_B) * np.cosh(ku * y)
                # print(Bo)
                f_base = np.cos
            else:  # coordonnee == 'z' :
                Bo = -(self.K * K_to_B) * np.sinh(ku * y)
                f_base = np.sin

            # we enelarge the real effect of the magnetic field by 4 lambda_u
            # on each extremity of the undulator
            L_magn_field = self.length / 2.0 + 4.0 * self.period_length

            # we know considere that if -(L/2 + lambda_u/4) < Z < (L/2 + lambda_u/4)
            # the magnetic field if a classic cosinus (or sinus)
            if coordonnee == 'y':
                a1 = self.length / self.period_length \
                     - np.floor(self.period_number())
                a2 = (0.25 - a1 / 2.0)
                L_cosinus_part = self.length / 2.0 \
                                 + self.period_length * a2

            else:
                L_cosinus_part = self.length / 2.0

            if ((z < -L_magn_field) or (z > L_magn_field)):
                dB = 0.0

            else:
                if (z < -L_cosinus_part or z > L_cosinus_part):

                    # in this case, we work with a gaussian,
                    # so we shift the coordinate frame for use a central gaussian
                    if z < -L_cosinus_part:
                        sign = 1
                    else:  # z> L_cosinus_part
                        sign = -1

                    shift_z = z + sign * L_cosinus_part

                    p = 2.0 * np.pi ** 2 / (3.0 * lambda_h ** 2)
                    dB = ((2.0 * np.pi * Bo / lambda_h) * shift_z) * (
                    1.0 - 4.0 * np.pi ** 2 * shift_z ** 2 / (9.0 * lambda_h ** 2)
                    ) * np.exp(-p * shift_z ** 2)

                    # test du signe
                    z_test = sign * (-L_cosinus_part + lambda_h / 4.0)
                    test_trig = f_base(ku * z_test)
                    if (sign * test_trig < 0):
                        dB = -dB

                else:
                    # in this case we work in the cosinus part
                    dB = Bo * f_base(ku * z)

        return dB


    def print_parameters(self):
        print(' Magnetic Structure (Plane Undulator)')
        print('    K : %.2f'%(self.K))
        print('    period lenght : %.3f' % (self.period_length))
        print('    lenght : %.2f'% (self.length))
        print('    number of periods : %d (L/lambdau=%.2f)' % (int(self.period_number()),self.period_number()))
        print('    magntic field intensity: %.2f'% (self.magnetic_field_strength()))

    def period_number(self):
        return self.length / self.period_length

    def magnetic_field_strength(self):
        Bo = (2.0 * np.pi * codata.m_e * codata.c * self.K) / (
            self.period_length * codata.e)
        return Bo

if __name__ == "__main__" :
    und_test = MagneticStructureUndulatorPlane(K=1.87, period_length=0.035, length=0.035 * 14)
    ESRF18 = MagneticStructureUndulatorPlane(K=1.68, period_length=0.018, length=2.0)

    print('undulator test :')
    und_test.print_parameters()
    B=und_test.create_magnetic_field()
    Z=np.linspace(-und_test.length/2.0,und_test.length/2.0,10000)
    B.plot_z(Z=Z,Y=0.0,X=0.0)

    print('undulator ESRF18 :')
    ESRF18.print_parameters()
    B=ESRF18.create_magnetic_field()
    Z=np.linspace(-ESRF18.length/2.0,ESRF18.length/2.0,10000)
    B.plot_z(Z=Z,Y=0.0,X=0.0)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import numpy as np
import scipy.constants as codata
from scipy.integrate import odeint
from pySRU.Trajectory import Trajectory
from pySRU.SourceUndulatorPlane import SourceUndulatorPlane,PLANE_UNDULATOR,BENDING_MAGNET
from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane as Undulator
from pySRU.ElectronBeam import ElectronBeam


TRAJECTORY_METHOD_ANALYTIC=0
TRAJECTORY_METHOD_ODE=1






def fct_ODE_magnetic_field(y, t, cst, Bx,By,Bz):
    return [cst * (Bz(z=y[5],y=y[4],x=y[3]) * y[1] - By(z=y[5],y=y[4],x=y[3]) * y[2]),
            cst * (Bx(z=y[5],y=y[4],x=y[3]) * y[2] - Bz(z=y[5],y=y[4],x=y[3]) * y[0]),
            cst * (By(z=y[5],y=y[4],x=y[3]) * y[0] - Bx(z=y[5],y=y[4],x=y[3]) * y[1]),
            y[0],
            y[1],
            y[2]]



'''
initial condition : [Vx,Vy,Vz,x,y,z]
'''
class TrajectoryFactory(object):
    def __init__(self,Nb_pts,method,initial_condition=None):
        self.Nb_pts = Nb_pts
        self.method = method
        self.initial_condition=initial_condition

    def copy(self):
        if self.initial_condition is None :
            cond=None
        else :
            cond= self.initial_condition.copy()
        return TrajectoryFactory(Nb_pts=self.Nb_pts,method=self.method,
                                 initial_condition=cond)

    def choise_initial_condition(self,source):
        if self.method != TRAJECTORY_METHOD_ANALYTIC:
            self.initial_condition = source.choose_initial_contidion_automatic()

    # calculate a theorical trajectory in an undulator
    def analytical_trajectory_plane_undulator(self,undulator):
        ku = 2.0 * np.pi / undulator.magnetic_structure.period_length
        gamma = undulator.Lorentz_factor()
        Beta_et = undulator.average_z_speed_in_undulator()
        K=undulator.magnetic_structure.K
        omega_u = Beta_et * codata.c * ku

        t = undulator.analytical_times_vector(Nb_pts=self.Nb_pts)
        ## x and z
        z = Beta_et * t + ((K / gamma) ** 2) * (1.0 / (8.0 * omega_u)) * np.sin( 2.0 * omega_u*t)
        x = (-(K / (gamma * omega_u)) * np.cos(omega_u*t))
        # # Vx and Vz
        v_z = Beta_et + ((K / gamma) ** 2) * (1.0 / 4.0) * np.cos(2.0 *omega_u*t)
        v_x= (K / (gamma )) * np.sin(omega_u*t)
        # # Ax and Az
        a_z=-omega_u *(K / gamma) ** 2 * 0.5 * np.sin( 2.0 * omega_u*t)
        a_x= (K / (gamma )) * (omega_u ) * np.cos(omega_u*t)
        # y
        y=0.0*t
        v_y=y
        a_y=y
        return Trajectory(t=t,x=x,y=y,z=z,v_x=v_x,v_y=v_y,v_z=v_z,
                                  a_x=a_x,a_y=a_y,a_z=a_z)



    # #TODO changer pour mettre avec des conditions initial ?
    def analytical_trajectory_cst_magnf(self, bending_magnet):
        Bo= bending_magnet.magnetic_structure.Bo
        #ro=bending_magnet.magnetic_structure.radius_curvature(bending_magnet.E())
        omega_p=Bo*codata.e/(codata.m_e*bending_magnet.Lorentz_factor())
        vz_0 = bending_magnet.electron_speed()

        t=bending_magnet.analytical_times_vector(Nb_pts=self.Nb_pts)
        Zo = bending_magnet.magnetic_structure.length / 2.
        #t=np.linspace(0.0,2.*Zo/(vz_0*codata.c))
        to=t[0]
        x = (vz_0/omega_p)*(1.-np.cos(omega_p*(t-to)))
        y = 0.0 * (t)
        z = (vz_0/omega_p)*np.sin(omega_p*(t-to))-Zo/codata.c
        vx = vz_0 * np.sin(omega_p*(t-to))
        vy = vz_0 * 0.0 * (t)
        vz = vz_0 * np.cos(omega_p*(t-to))
        ax = vz_0 * omega_p * np.cos(omega_p*(t-to))
        ay = vz_0 * 0.0 * (t)
        az = -vz_0* omega_p * np.sin(omega_p*(t-to))

        return Trajectory(t=t,x=x,y=y,z=z,v_x=vx,v_y=vy,v_z=vz,a_x=ax,a_y=ay,a_z=az)




    def trajectory_from_magnetic_field_method_ODE(self, source,t=None):
        gamma = source.Lorentz_factor()

        B=source.magnetic_field
        #   trajectory =
        #   [t........]
        # 	[ X/c......]
        # 	[ Y/c ......]
        #	[ Z/c ......]
        # 	[ Vx/c .....]
        # 	[ Vy/c .....]
        # 	[ Vz/c .....]
        # 	[ Ax/c .....]
        # 	[ Ay/c .....]
        # 	[ Az/c .....]

        if t is None :
            time_calc = source.construct_times_vector(initial_contition=self.initial_condition,Nb_pts=self.Nb_pts)
        else :
            self.Nb_pts=len(t)
            time_calc =t

        trajectory = np.zeros((10, self.Nb_pts))
        trajectory[0]=time_calc

        cst = -codata.e / (codata.m_e * gamma)
        initial_condition_for_ODE=self.initial_condition

        #TODO rtol et a tol modifiable
        #atol=np.array([1e-10,1e-10,1e-10,1e-10,1e-10,1e-10])
        rtol=source.rtol_for_ODE_method()
        atol=source.atol_for_ODE_method()
        res = odeint(func=fct_ODE_magnetic_field,y0=initial_condition_for_ODE, t=trajectory[0],
                     args=(cst,B.Bx,B.By,B.Bz),rtol=rtol,atol=atol,mxstep=5000,full_output=True)
        traj = res[0]
        info = res[1]
        print("1 : nonstiff problems, Adams . 2: stiff problem, BDF")
        print(info.get('mused'))
        traj = np.transpose(traj)
        trajectory[4] = traj[0]
        trajectory[5] = traj[1]
        trajectory[6] = traj[2]
        trajectory[1] = traj[3]
        trajectory[2] = traj[4]
        trajectory[3] = traj[5]
        trajectory[7] = - cst * B.By(trajectory[3], trajectory[2],trajectory[1]) * trajectory[6]
        trajectory[9] = cst * B.By(trajectory[3], trajectory[2],trajectory[1]) * trajectory[4]
        T=self.create_from_array(trajectory)
        T.multiply_by((1.0/codata.c))
        return T


    def create_from_source(self, source):
        if (self.method == TRAJECTORY_METHOD_ODE):
            if (self.initial_condition is None):
                self.choise_initial_condition(source=source)
            trajectory = self.trajectory_from_magnetic_field_method_ODE(source=source)
        else:
            if source.magnet_type()==PLANE_UNDULATOR:
                trajectory = self.analytical_trajectory_plane_undulator(undulator=source)
            else :
                trajectory = self.analytical_trajectory_cst_magnf(bending_magnet=source)
            self.initial_condition = np.array([trajectory.v_x[0], trajectory.v_y[0], trajectory.v_z[0],
                                                   trajectory.x[0], trajectory.y[0], trajectory.z[0]])
            self.initial_condition *= codata.c

        return trajectory

    def create_from_array(self,array):
        if array.shape[0] != 10 :
            raise Exception('this array can not be convert in Trajectory')
        return Trajectory(t=array[0],x=array[1],y=array[2],z=array[3],v_x=array[4],v_y=array[5],
                    v_z = array[6],a_x=array[7],a_y=array[8],a_z=array[9])

    def get_method(self):
        if self.method==TRAJECTORY_METHOD_ANALYTIC :
            method='Analitical Trajectory'

        elif self.method == TRAJECTORY_METHOD_ODE:
            method = ' ODE solution '

        else :# self.method == TRAJECTORY_METHOD_Integration:
            method = ' Trajectory from integration on the magnetic field'
        return method


    def print_parameters(self):
        print("Trajectory ")
        print('    method : %s' %self.get_method())
        print('    number of points : %d' %self.Nb_pts)
        print('    initial position (x,y,z) : ')
        print(self.initial_condition[3:6])
        print('    initial velocity (x,y,z)  ')
        print(self.initial_condition[0:3])


if __name__ == "__main__" :
    from SourceUndulatorPlane import SourceUndulatorPlane
    undulator_test = Undulator(K=1.87, period_length=0.035, length=0.035 * 14)
    electron_beam_test = ElectronBeam(Electron_energy=1.3e9, I_current=1.0)

    source_test=SourceUndulatorPlane(undulator=undulator_test,electron_beam=electron_beam_test)


    print('Create trajectory with autamatic choice of initial condition and automatic magnetic field')

    trajectory_fact_ODE = TrajectoryFactory(Nb_pts=20000, method=TRAJECTORY_METHOD_ODE)
    trajectory1=trajectory_fact_ODE.create_from_source(source_test)
    print(' ')
    print('trajectory 1 created with ODE method')
    print(trajectory_fact_ODE.print_parameters())
    print('initial condition =')
    print(trajectory_fact_ODE.initial_condition * (1. / codata.c))
    trajectory1.plot_3D(title="traj 1: autamatic initial condition and field")

    trajectory_fact_ANALITIC = TrajectoryFactory(Nb_pts=20000, method=TRAJECTORY_METHOD_ANALYTIC)
    trajectory2=trajectory_fact_ANALITIC.create_from_source(source_test)
    print(' ')
    print('trajectory 2 create with ANALYTIC method')
    print(trajectory_fact_ANALITIC.print_parameters())
    print('initial condition =')
    print(trajectory_fact_ANALITIC.initial_condition*(1./codata.c))
    trajectory2.plot_3D(title="traj 2: ANALYTICAL method")


    print(' ')
    print('initial condition can be impose:')
    print('trajectory 1 modified with initial condition of trajectory2')
    trajectory_fact_ODE.initial_condition=trajectory_fact_ANALITIC.initial_condition
    #print(all(trajectory_fact_ODE.initial_condition==trajectory_fact_ANALITIC.initial_condition))
    trajectory1=trajectory_fact_ODE.create_from_source(source_test)
    trajectory1.plot_3D("traj 1 modified with intitial conditions from traj 2")

    print(' ')
    print('Now trajectory 1 and 2 have the same vector time, we make the difference : ')
    diff =trajectory1.difference_with(trajectory2)

    diff.plot(title="Difference")




# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["S Thery, M Glass, M Sanchez del Rio - ESRF ISDD Advanced Analysis and Modelling"]
__license__ = "MIT"
__date__ = "31/08/2016"

import unittest
import numpy as np
from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane as Undulator
from pySRU.ElectronBeam import ElectronBeam
from pySRU.SourceUndulatorPlane import SourceUndulatorPlane


class UndulatorParameterTest(unittest.TestCase):

    def test_copy(self):
        K = 1.87
        lambda_u = 0.035
        L = 0.035 * 14
        E=1.3
        I=1.0
        undulator = Undulator(K=K, period_length=lambda_u, length=L)
        beam=ElectronBeam(Electron_energy=E, I_current=I)
        source=SourceUndulatorPlane(undulator=undulator, electron_beam=beam)

        source2=source.copy()
        source2.electron_beam.I_current=0.3
        self.assertEqual(source.I_current(),1.0)
        self.assertEqual(source2.I_current() ,0.3)

        self.assertAlmostEqual(source.harmonic_frequency(1)/1e17, 2.5346701615509917,5)
        self.assertAlmostEqual(source.Lorentz_factor(), 2544.0367765521196,2 )
        self.assertAlmostEqual(source.electron_speed(), 0.99999992274559524,5)
        self.assertEqual(source.magnetic_structure.period_number(), 14)
        self.assertAlmostEqual(source.choose_distance_automatic(2),49.000000000000,10 )



    #TODO completer ?
